import pytest
from unittest import TestCase
from onyx import OnyxConfig
from onyx.exceptions import OnyxConfigError


DOMAIN = "https://onyx.domain"
TOKEN = "token"
USERNAME = "username"
PASSWORD = "password"


class OnyxConfigTestCase(TestCase):
    def test_init(self):
        config = OnyxConfig(
            domain=DOMAIN,
            token=TOKEN,
            username=USERNAME,
            password=PASSWORD,
        )
        self.assertEqual(config.domain, DOMAIN)
        self.assertEqual(config.token, TOKEN)
        self.assertEqual(config.username, USERNAME)
        self.assertEqual(config.password, PASSWORD)

        config = OnyxConfig(
            domain=DOMAIN,
            token=TOKEN,
        )
        self.assertEqual(config.domain, DOMAIN)
        self.assertEqual(config.token, TOKEN)

        config = OnyxConfig(
            domain=DOMAIN,
            username=USERNAME,
            password=PASSWORD,
        )
        self.assertEqual(config.domain, DOMAIN)
        self.assertEqual(config.username, USERNAME)
        self.assertEqual(config.password, PASSWORD)

        # TODO: Handle " "
        for empty in ["", None]:
            with pytest.raises(OnyxConfigError):
                OnyxConfig(
                    domain=empty,
                    token=TOKEN,
                    username=USERNAME,
                    password=PASSWORD,
                )

            with pytest.raises(OnyxConfigError):
                OnyxConfig(
                    domain=empty,
                    token=TOKEN,
                )

            with pytest.raises(OnyxConfigError):
                OnyxConfig(
                    domain=empty,
                    username=USERNAME,
                    password=PASSWORD,
                )

            with pytest.raises(OnyxConfigError):
                OnyxConfig(
                    domain=DOMAIN,
                    token=empty,
                    username=empty,
                    password=empty,
                )

            with pytest.raises(OnyxConfigError):
                OnyxConfig(
                    domain=DOMAIN,
                    token=empty,
                    username=USERNAME,
                    password=empty,
                )

            with pytest.raises(OnyxConfigError):
                OnyxConfig(
                    domain=DOMAIN,
                    token=empty,
                    username=empty,
                    password=PASSWORD,
                )



import io
import requests
import pytest
from unittest import TestCase, mock
from onyx import OnyxConfig, OnyxClient, exceptions, OnyxField


DOMAIN = "https://onyx.domain"
BAD_DOMAIN = "not-onyx"
TOKEN = "token"
EXPIRY = "expiry"
FIRST_NAME = "first_name"
LAST_NAME = "last_name"
USERNAME = "username"
PASSWORD = "password"
EMAIL = "email"
SITE = "site"
OTHER_SITE = "other_site"
OTHER_USERNAME = "other_username"
OTHER_EMAIL = "other_email"
INVALID_AUTH_DATA = {
    "status": "fail",
    "code": 401,
    "messages": {
        "detail": "Invalid username/password.",
    },
}
PROJECT = "project"
NOT_PROJECT = "not_project"
ERROR_CAUSING_PROJECT = "error_causing_project"
PROJECT_DATA = {
    "status": "success",
    "code": 200,
    "data": [
        {
            "project": PROJECT,
            "action": "view",
            "scope": "base",
        }
    ],
}
TYPES_DATA = {
    "status": "success",
    "code": 200,
    "data": [
        {
            "type": "text",
            "description": "A text field.",
            "lookups": [
                "exact",
                "iexact",
                "contains",
                "icontains",
                "startswith",
                "istartswith",
                "endswith",
                "iendswith",
            ],
        },
        {
            "type": "choice",
            "description": "A choice field.",
            "lookups": [
                "exact",
                "iexact",
            ],
        },
    ],
}
LOOKUPS_DATA = {
    "status": "success",
    "code": 200,
    "data": [
        {
            "lookup": "exact",
            "description": "Exact match.",
            "types": [
                "text",
                "choice",
            ],
        },
        {
            "lookup": "iexact",
            "description": "Case-insensitive exact match.",
            "types": [
                "text",
                "choice",
            ],
        },
    ],
}
FIELDS_DATA = {
    "status": "success",
    "code": 200,
    "data": {
        "version": "0.1.0",
        "fields": {
            "climb_id": {
                "type": "text",
                "required": True,
                "description": "Unique identifier for a project. Set by Onyx.",
            },
            "sample_id": {
                "type": "text",
                "required": True,
                "description": "Unique identifier for a biological sample.",
            },
            "run_name": {
                "type": "text",
                "required": True,
                "description": "Unique identifier for a sequencing run.",
            },
            "source_type": {
                "type": "text",
                "required": True,
                "description": "Where was the sample sourced. Is it human, animal... or something else...",
            },
        },
    },
}
FIELDS_NOT_PROJECT_DATA = {
    "status": "fail",
    "code": 404,
    "messages": {
        "detail": "Project not found.",
    },
}
FIELDS_ERROR_CAUSING_PROJECT_DATA = {
    "status": "fail",
    "code": 500,
    "messages": {
        "detail": "Internal server error. Oh dear...",
    },
}
CHOICE_FIELD = "country"
CHOICES_DATA = {
    "status": "success",
    "code": 200,
    "data": [
        {
            "eng": {
                "description": "England",
                "is_active": True,
            },
            "ni": {
                "description": "N. Ireland",
                "is_active": True,
            },
            "scot": {
                "description": "Scotland",
                "is_active": True,
            },
            "wales": {
                "description": "Wales",
                "is_active": True,
            },
        }
    ],
}
CLIMB_ID = "C-0123456789"
GET_DATA = {
    "status": "success",
    "code": 200,
    "data": {
        "climb_id": CLIMB_ID,
        "published_date": "2023-09-18",
        "sample_id": "sample-123",
        "run_name": "run-456",
    },
}
FILTER_PAGE_1_URL = f"{OnyxClient.ENDPOINTS['filter'](DOMAIN, PROJECT)}?cursor=page_1"
FILTER_PAGE_2_URL = f"{OnyxClient.ENDPOINTS['filter'](DOMAIN, PROJECT)}?cursor=page_2"
FILTER_PAGE_1_DATA = {
    "status": "success",
    "code": 200,
    "next": FILTER_PAGE_2_URL,
    "previous": None,
    "data": [
        {
            "climb_id": CLIMB_ID,
            "published_date": "2023-09-18",
            "sample_id": "sample-123",
            "run_name": "run-456",
        },
        {
            "climb_id": CLIMB_ID,
            "published_date": "2023-09-18",
            "sample_id": "sample-123",
            "run_name": "run-456",
        },
        {
            "climb_id": CLIMB_ID,
            "published_date": "2023-09-18",
            "sample_id": "sample-123",
            "run_name": "run-456",
        },
    ],
}
FILTER_PAGE_2_DATA = {
    "status": "success",
    "code": 200,
    "next": None,
    "previous": FILTER_PAGE_1_URL,
    "data": [
        {
            "climb_id": CLIMB_ID,
            "published_date": "2023-09-18",
            "sample_id": "sample-123",
            "run_name": "run-456",
        },
        {
            "climb_id": CLIMB_ID,
            "published_date": "2023-09-18",
            "sample_id": "sample-123",
            "run_name": "run-456",
        },
        {
            "climb_id": CLIMB_ID,
            "published_date": "2023-09-18",
            "sample_id": "sample-123",
            "run_name": "run-456",
        },
    ],
}
SAMPLE_ID = "sample-abc"
RUN_NAME = "run-def"
PUBLISHED_DATE_RANGE = ["2023-01-01", "2024-01-01"]
FILTER_SPECIFIC_DATA = {
    "status": "success",
    "code": 200,
    "next": None,
    "previous": None,
    "data": [
        {
            "climb_id": CLIMB_ID,
            "published_date": "2023-09-18",
            "sample_id": "sample-abc",
            "run_name": "run-def",
        },
    ],
}
INCLUDE_FIELDS = ["climb_id", "published_date"]
FILTER_SPECIFIC_INCLUDE_DATA = {
    "status": "success",
    "code": 200,
    "next": None,
    "previous": None,
    "data": [
        {
            "climb_id": CLIMB_ID,
            "published_date": "2023-09-18",
        },
    ],
}
EXCLUDE_FIELDS = ["run_name"]
FILTER_SPECIFIC_EXCLUDE_DATA = {
    "status": "success",
    "code": 200,
    "next": None,
    "previous": None,
    "data": [
        {
            "climb_id": CLIMB_ID,
            "published_date": "2023-09-18",
            "sample_id": "sample-abc",
        },
    ],
}
UNKNOWN_FIELD = "haha"
NONE_FIELD = "field-that-has-none-values"
FILTER_NONE_DATA = {
    "status": "success",
    "code": 200,
    "next": None,
    "previous": None,
    "data": [
        {
            "climb_id": CLIMB_ID,
            NONE_FIELD: None,
        }
    ],
}
FILTER_NONE_IN_DATA = {
    "status": "success",
    "code": 200,
    "next": None,
    "previous": None,
    "data": [
        {
            "climb_id": CLIMB_ID,
            NONE_FIELD: None,
        },
        {
            "climb_id": CLIMB_ID,
            NONE_FIELD: "not-none",
        },
    ],
}
FILTER_UNKNOWN_FIELD_DATA = {
    "status": "fail",
    "code": 400,
    "messages": {UNKNOWN_FIELD: ["This field is unknown."]},
}
FILTER_ERROR_CAUSING_PROJECT_DATA = {
    "status": "fail",
    "code": 500,
    "messages": {
        "detail": "Internal server error. Deary me...",
    },
}
QUERY_PAGE_1_URL = f"{OnyxClient.ENDPOINTS['query'](DOMAIN, PROJECT)}?cursor=page_1"
QUERY_PAGE_2_URL = f"{OnyxClient.ENDPOINTS['query'](DOMAIN, PROJECT)}?cursor=page_2"
QUERY_PAGE_1_DATA = {
    "status": "success",
    "code": 200,
    "next": QUERY_PAGE_2_URL,
    "previous": None,
    "data": [
        {
            "climb_id": CLIMB_ID,
            "published_date": "2023-09-18",
            "sample_id": "sample-123",
            "run_name": "run-456",
        },
        {
            "climb_id": CLIMB_ID,
            "published_date": "2023-09-18",
            "sample_id": "sample-123",
            "run_name": "run-456",
        },
        {
            "climb_id": CLIMB_ID,
            "published_date": "2023-09-18",
            "sample_id": "sample-123",
            "run_name": "run-456",
        },
    ],
}
QUERY_PAGE_2_DATA = {
    "status": "success",
    "code": 200,
    "next": None,
    "previous": QUERY_PAGE_1_URL,
    "data": [
        {
            "climb_id": CLIMB_ID,
            "published_date": "2023-09-18",
            "sample_id": "sample-123",
            "run_name": "run-456",
        },
        {
            "climb_id": CLIMB_ID,
            "published_date": "2023-09-18",
            "sample_id": "sample-123",
            "run_name": "run-456",
        },
        {
            "climb_id": CLIMB_ID,
            "published_date": "2023-09-18",
            "sample_id": "sample-123",
            "run_name": "run-456",
        },
    ],
}
QUERY_SPECIFIC_BODY = {"&": [{"sample_id": SAMPLE_ID}, {"run_name": RUN_NAME}]}
HISTORY_DATA = {
    "status": "success",
    "code": 200,
    "data": {
        "climb_id": CLIMB_ID,
        "history": [
            {
                "username": USERNAME,
                "timestamp": "2024-01-01T12:00:00Z",
                "action": "add",
            },
            {
                "username": USERNAME,
                "timestamp": "2024-01-01T12:00:01Z",
                "action": "change",
                "changes": [],
            },
        ],
    },
}
IDENTIFY_FIELD = "sample_id"
IDENTIFY_VALUE = "sample-123"
IDENTIFY_FIELDS = {"value": IDENTIFY_VALUE}
IDENTIFY_DATA = {
    "status": "success",
    "code": 200,
    "data": {
        "project": PROJECT,
        "site": SITE,
        "field": IDENTIFY_FIELD,
        "value": IDENTIFY_VALUE,
        "identifier": "S-1234567890",
    },
}
IDENTIFY_OTHER_SITE_FIELDS = {"value": IDENTIFY_VALUE, "site": OTHER_SITE}
IDENTIFY_OTHER_SITE_DATA = {
    "status": "success",
    "code": 200,
    "data": {
        "project": PROJECT,
        "site": OTHER_SITE,
        "field": IDENTIFY_FIELD,
        "value": IDENTIFY_VALUE,
        "identifier": "S-0987654321",
    },
}
CREATE_FIELDS = {
    "sample_id": "sample-123",
    "run_name": "run-456",
}
CSV_CREATE_EMPTY_FILE = "sample_id, run_name\n"
TSV_CREATE_EMPTY_FILE = "sample_id\t run_name\n"
CSV_CREATE_SINGLE_FILE = "sample_id, run_name\nsample-123, run-456"
TSV_CREATE_SINGLE_FILE = "sample_id\t run_name\nsample-123\t run-456"
CSV_CREATE_MULTI_FILE = "sample_id, run_name\nsample-123, run-456\nsample-123, run-456"
TSV_CREATE_MULTI_FILE = (
    "sample_id\t run_name\nsample-123\t run-456\nsample-123\t run-456"
)
CSV_CREATE_SINGLE_MISSING_FILE = "sample_id\nsample-123"
TSV_CREATE_SINGLE_MISSING_FILE = "sample_id\nsample-123"
CSV_CREATE_MULTI_MISSING_FILE = "sample_id\nsample-123\nsample-123"
TSV_CREATE_MULTI_MISSING_FILE = "sample_id\nsample-123\nsample-123"
MISSING_CREATE_FIELDS = {"run_name": "run-456"}
CREATE_DATA = {
    "status": "success",
    "code": 201,
    "data": {
        "climb_id": CLIMB_ID,
    },
}
TESTCREATE_DATA = {
    "status": "success",
    "code": 201,
    "data": {
        "climb_id": None,
    },
}
UPDATE_FIELDS = {
    "country": "England",
    "source_type": "humanoid",
}
CSV_UPDATE_EMPTY_FILE = "climb_id, country, source_type\n"
TSV_UPDATE_EMPTY_FILE = "climb_id\t country\t source_type\n"
CSV_UPDATE_SINGLE_FILE = (
    f"climb_id, country, source_type\n{CLIMB_ID}, England, humanoid"
)
TSV_UPDATE_SINGLE_FILE = (
    f"climb_id\t country\t source_type\n{CLIMB_ID}\t England\t humanoid"
)
CSV_UPDATE_MULTI_FILE = f"climb_id, country, source_type\n{CLIMB_ID}, England, humanoid\n{CLIMB_ID}, England, humanoid"
TSV_UPDATE_MULTI_FILE = f"climb_id\t country\t source_type\n{CLIMB_ID}\t England\t humanoid\n{CLIMB_ID}\t England\t humanoid"
CSV_UPDATE_SINGLE_MISSING_FILE = "country, source_type\nEngland, humanoid"
TSV_UPDATE_SINGLE_MISSING_FILE = "country\t source_type\nEngland\t humanoid"
CSV_UPDATE_MULTI_MISSING_FILE = (
    "country, source_type\nEngland, humanoid\nEngland, humanoid"
)
TSV_UPDATE_MULTI_MISSING_FILE = (
    "country\t source_type\nEngland\t humanoid\nEngland\t humanoid"
)
MISSING_UPDATE_FIELDS = {"climb_id": CLIMB_ID}
UPDATE_DATA = {
    "status": "success",
    "code": 200,
    "data": {
        "climb_id": CLIMB_ID,
    },
}
TESTUPDATE_DATA = {
    "status": "success",
    "code": 200,
    "data": {
        "climb_id": CLIMB_ID,
    },
}
CSV_DELETE_EMPTY_FILE = f"climb_id\n"
TSV_DELETE_EMPTY_FILE = f"climb_id\n"
CSV_DELETE_SINGLE_FILE = f"climb_id\n{CLIMB_ID}"
TSV_DELETE_SINGLE_FILE = f"climb_id\n{CLIMB_ID}"
CSV_DELETE_MULTI_FILE = f"climb_id\n{CLIMB_ID}\n{CLIMB_ID}"
TSV_DELETE_MULTI_FILE = f"climb_id\n{CLIMB_ID}\n{CLIMB_ID}"
DELETE_DATA = {
    "status": "success",
    "code": 200,
    "data": {
        "climb_id": CLIMB_ID,
    },
}
REGISTER_FIELDS = {
    "first_name": FIRST_NAME,
    "last_name": LAST_NAME,
    "email": EMAIL,
    "password": PASSWORD,
    "site": SITE,
}
REGISTER_DATA = {
    "status": "success",
    "code": 201,
    "data": {
        "username": USERNAME,
        "site": SITE,
        "email": EMAIL,
        "first_name": FIRST_NAME,
        "last_name": LAST_NAME,
    },
}
LOGIN_DATA = {
    "status": "success",
    "code": 200,
    "data": {
        "token": TOKEN,
        "expiry": EXPIRY,
    },
}
LOGOUT_DATA = {
    "status": "success",
    "code": 204,
    "data": None,
}
LOGOUTALL_DATA = {
    "status": "success",
    "code": 204,
    "data": None,
}
PROFILE_DATA = {
    "status": "success",
    "code": 200,
    "data": {
        "username": USERNAME,
        "email": EMAIL,
        "site": SITE,
    },
}
ACTIVITY_DATA = {
    "status": "success",
    "code": 200,
    "data": [
        {
            "date": "2023-01-01T00:00:00.000000Z",
            "address": "127.0.0.1",
            "endpoint": "/projects/project/",
            "method": "POST",
            "status": 400,
            "exec_time": 29,
            "error_messages": 'b\'{"status":"fail","code":400,"messages":{"site":["Select a valid choice."]}}\'',
        },
        {
            "timestamp": "2023-01-02T00:00:00.000000Z",
            "address": "127.0.0.1",
            "endpoint": "/accounts/activity/",
            "method": "GET",
            "status": 200,
            "exec_time": 22,
            "error_messages": "",
        },
    ],
}
WAITING_DATA = {
    "status": "success",
    "code": 200,
    "data": [
        {
            "username": OTHER_USERNAME,
            "email": OTHER_EMAIL,
            "site": SITE,
        }
    ],
}
SITE_USERS_DATA = {
    "status": "success",
    "code": 200,
    "data": [
        {
            "username": USERNAME,
            "email": EMAIL,
            "site": SITE,
        }
    ],
}
ALL_USERS_DATA = {
    "status": "success",
    "code": 200,
    "data": [
        {
            "username": USERNAME,
            "email": EMAIL,
            "site": SITE,
        },
        {
            "username": OTHER_USERNAME,
            "email": OTHER_EMAIL,
            "site": SITE,
        },
    ],
}
APPROVE_DATA = {
    "status": "success",
    "code": 200,
    "data": {
        "username": OTHER_USERNAME,
        "is_approved": True,
    },
}
INVALID_DOMAIN_ARGUMENTS = ["", " ", None]
INVALID_ARGUMENTS = INVALID_DOMAIN_ARGUMENTS + ["/", "?", "/?"]
PROJECT_ENDPOINT_CLASHES = ["types", "lookups"]
CLIMB_ID_ENDPOINT_CLASHES = [
    "test",
    "query",
    "fields",
    "choices",
    "history",
    "identify",
]


class MockResponse:
    def __init__(self, data):
        self.data = data
        self.status_code = data["code"]
        if self.status_code < 400:
            self.ok = True
        else:
            self.ok = False

    def json(self):
        return self.data

    def raise_for_status(self):
        if self.status_code >= 400:
            raise requests.HTTPError(
                "Something terrible happened.",
                response=self,  #  type: ignore
            )


def mock_request(
    method=None,
    headers=None,
    auth=None,
    url=None,
    params=None,
    json=None,
):
    if not headers:
        headers = {}

    if not url:
        url = ""

    if not params:
        params = {}

    if not json:
        json = {}

    if url.startswith(BAD_DOMAIN):
        raise requests.ConnectionError

    if method == "post" and url == OnyxClient.ENDPOINTS["login"](DOMAIN):
        if auth == (USERNAME, PASSWORD):
            return MockResponse(LOGIN_DATA)
        else:
            return MockResponse(INVALID_AUTH_DATA)

    if headers.get("Authorization") != f"Token {TOKEN}":
        return MockResponse(INVALID_AUTH_DATA)

    if method == "post":
        if (
            url == OnyxClient.ENDPOINTS["create"](DOMAIN, PROJECT)
            and json == CREATE_FIELDS
        ):
            return MockResponse(CREATE_DATA)

        elif (
            url == OnyxClient.ENDPOINTS["testcreate"](DOMAIN, PROJECT)
            and json == CREATE_FIELDS
        ):
            return MockResponse(TESTCREATE_DATA)

        elif url == OnyxClient.ENDPOINTS["query"](DOMAIN, PROJECT):
            if json == QUERY_SPECIFIC_BODY:
                if params.get("include") == INCLUDE_FIELDS:
                    return MockResponse(FILTER_SPECIFIC_INCLUDE_DATA)
                elif params.get("exclude") == EXCLUDE_FIELDS:
                    return MockResponse(FILTER_SPECIFIC_EXCLUDE_DATA)
                else:
                    return MockResponse(FILTER_SPECIFIC_DATA)
            else:
                return MockResponse(QUERY_PAGE_1_DATA)

        elif url == QUERY_PAGE_2_URL:
            return MockResponse(QUERY_PAGE_2_DATA)

        elif url == OnyxClient.ENDPOINTS["logout"](DOMAIN):
            return MockResponse(LOGOUT_DATA)

        elif url == OnyxClient.ENDPOINTS["logoutall"](DOMAIN):
            return MockResponse(LOGOUTALL_DATA)

        elif url == OnyxClient.ENDPOINTS["identify"](DOMAIN, PROJECT, IDENTIFY_FIELD):
            if json == IDENTIFY_FIELDS:
                return MockResponse(IDENTIFY_DATA)

            elif json == IDENTIFY_OTHER_SITE_FIELDS:
                return MockResponse(IDENTIFY_OTHER_SITE_DATA)

    elif method == "get":
        if url == OnyxClient.ENDPOINTS["projects"](DOMAIN):
            return MockResponse(PROJECT_DATA)

        elif url == OnyxClient.ENDPOINTS["types"](DOMAIN):
            return MockResponse(TYPES_DATA)

        elif url == OnyxClient.ENDPOINTS["lookups"](DOMAIN):
            return MockResponse(LOOKUPS_DATA)

        elif url == OnyxClient.ENDPOINTS["fields"](DOMAIN, PROJECT):
            return MockResponse(FIELDS_DATA)

        elif url == OnyxClient.ENDPOINTS["fields"](DOMAIN, NOT_PROJECT):
            return MockResponse(FIELDS_NOT_PROJECT_DATA)

        elif url == OnyxClient.ENDPOINTS["fields"](DOMAIN, ERROR_CAUSING_PROJECT):
            return MockResponse(FIELDS_ERROR_CAUSING_PROJECT_DATA)

        elif url == OnyxClient.ENDPOINTS["choices"](DOMAIN, PROJECT, CHOICE_FIELD):
            return MockResponse(CHOICES_DATA)

        elif url == OnyxClient.ENDPOINTS["get"](DOMAIN, PROJECT, CLIMB_ID):
            return MockResponse(GET_DATA)

        elif url == OnyxClient.ENDPOINTS["filter"](DOMAIN, PROJECT):
            if params.get(UNKNOWN_FIELD):
                return MockResponse(FILTER_UNKNOWN_FIELD_DATA)
            elif params.get(NONE_FIELD) == "":
                return MockResponse(FILTER_NONE_DATA)
            elif "" in params.get(f"{NONE_FIELD}__in", []):
                return MockResponse(FILTER_NONE_IN_DATA)
            elif (
                params.get("sample_id") == SAMPLE_ID
                and params.get("run_name") == RUN_NAME
                and params.get("published_date__range")
                == ",".join(PUBLISHED_DATE_RANGE)
            ):
                if params.get("include") == INCLUDE_FIELDS:
                    return MockResponse(FILTER_SPECIFIC_INCLUDE_DATA)
                elif params.get("exclude") == EXCLUDE_FIELDS:
                    return MockResponse(FILTER_SPECIFIC_EXCLUDE_DATA)
                else:
                    return MockResponse(FILTER_SPECIFIC_DATA)
            else:
                return MockResponse(FILTER_PAGE_1_DATA)

        elif url == OnyxClient.ENDPOINTS["filter"](DOMAIN, ERROR_CAUSING_PROJECT):
            return MockResponse(FILTER_ERROR_CAUSING_PROJECT_DATA)

        elif url == FILTER_PAGE_2_URL:
            return MockResponse(FILTER_PAGE_2_DATA)

        elif url == OnyxClient.ENDPOINTS["history"](DOMAIN, PROJECT, CLIMB_ID):
            return MockResponse(HISTORY_DATA)

        elif url == OnyxClient.ENDPOINTS["profile"](DOMAIN):
            return MockResponse(PROFILE_DATA)

        elif url == OnyxClient.ENDPOINTS["activity"](DOMAIN):
            return MockResponse(ACTIVITY_DATA)

        elif url == OnyxClient.ENDPOINTS["waiting"](DOMAIN):
            return MockResponse(WAITING_DATA)

        elif url == OnyxClient.ENDPOINTS["siteusers"](DOMAIN):
            return MockResponse(SITE_USERS_DATA)

        elif url == OnyxClient.ENDPOINTS["allusers"](DOMAIN):
            return MockResponse(ALL_USERS_DATA)

    elif method == "patch":
        if (
            url == OnyxClient.ENDPOINTS["update"](DOMAIN, PROJECT, CLIMB_ID)
            and json == UPDATE_FIELDS
        ):
            return MockResponse(UPDATE_DATA)

        elif (
            url == OnyxClient.ENDPOINTS["testupdate"](DOMAIN, PROJECT, CLIMB_ID)
            and json == UPDATE_FIELDS
        ):
            return MockResponse(TESTUPDATE_DATA)

        elif url == OnyxClient.ENDPOINTS["approve"](DOMAIN, OTHER_USERNAME):
            return MockResponse(APPROVE_DATA)

    elif method == "delete":
        if url == OnyxClient.ENDPOINTS["delete"](DOMAIN, PROJECT, CLIMB_ID):
            return MockResponse(DELETE_DATA)

    return MockResponse(
        {
            "status": "fail",
            "code": 400,
            "messages": {"detail": "Something terrible happened."},
        },
    )


def mock_register_post(url=None, json=None):
    if not json:
        json = {}

    if url == OnyxClient.ENDPOINTS["register"](DOMAIN) and all(
        json.get(field) == value for field, value in REGISTER_FIELDS.items()
    ):
        return MockResponse(REGISTER_DATA)

    return MockResponse(
        {
            "status": "fail",
            "code": 400,
            "messages": {"detail": "Something terrible happened."},
        },
    )


class OnyxClientTestCase(TestCase):
    def setUp(self) -> None:
        self.config = OnyxConfig(
            domain=DOMAIN,
            username=USERNAME,
            password=PASSWORD,
        )
        self.client = OnyxClient(self.config)

    @mock.patch("onyx.OnyxClient._request_handler", side_effect=mock_request)
    def test_context_manager(self, mock_request):
        """
        Test that the OnyxClient can be used as a context manager.
        """

        with OnyxClient(self.config) as client:
            self.assertIsInstance(client._session, requests.Session)
            self.assertEqual(
                client._request_handler, client._session.request  #  type: ignore
            )

        self.assertEqual(client._request_handler, requests.request)

    @mock.patch("onyx.OnyxClient._request_handler", side_effect=mock_request)
    def test_connection_error(self, mock_request):
        """
        Test that the OnyxClient raises an OnyxConnectionError when a connection error occurs.
        """

        self.config.domain = BAD_DOMAIN

        # Non-generator connection error
        with pytest.raises(exceptions.OnyxConnectionError):
            self.client.projects()

        # Generator connection error
        with pytest.raises(exceptions.OnyxConnectionError):
            [x for x in self.client.filter(PROJECT)]

        self.assertEqual(self.config.token, None)

    @mock.patch("onyx.OnyxClient._request_handler", side_effect=mock_request)
    def test_request_error(self, mock_request):
        """
        Test that the OnyxClient raises an OnyxRequestError when a request error occurs.
        """

        # Non-generator request error
        with pytest.raises(exceptions.OnyxRequestError) as e:
            self.client.fields(NOT_PROJECT)
        self.assertEqual(e.value.response.json(), FIELDS_NOT_PROJECT_DATA)

        # Generator request error
        with pytest.raises(exceptions.OnyxRequestError) as e:
            [x for x in self.client.filter(PROJECT, fields={UNKNOWN_FIELD: "haha"})]
        self.assertEqual(e.value.response.json(), FILTER_UNKNOWN_FIELD_DATA)

        self.assertEqual(self.config.token, TOKEN)

    @mock.patch("onyx.OnyxClient._request_handler", side_effect=mock_request)
    def test_server_error(self, mock_request):
        """
        Test that the OnyxClient raises an OnyxServerError when a server error occurs.
        """

        # Non-generator server error
        with pytest.raises(exceptions.OnyxServerError) as e:
            self.client.fields(ERROR_CAUSING_PROJECT)
        self.assertEqual(e.value.response.json(), FIELDS_ERROR_CAUSING_PROJECT_DATA)

        # Generator server error
        with pytest.raises(exceptions.OnyxServerError) as e:
            [x for x in self.client.filter(ERROR_CAUSING_PROJECT)]
        self.assertEqual(e.value.response.json(), FILTER_ERROR_CAUSING_PROJECT_DATA)

        self.assertEqual(self.config.token, TOKEN)

    @mock.patch("onyx.OnyxClient._request_handler", side_effect=mock_request)
    def test_projects(self, mock_request):
        """
        Test that the OnyxClient can retrieve a list of projects.
        """

        self.assertEqual(self.client.projects(), PROJECT_DATA["data"])
        self.assertEqual(self.config.token, TOKEN)

    @mock.patch("onyx.OnyxClient._request_handler", side_effect=mock_request)
    def test_types(self, mock_request):
        """
        Test that the OnyxClient can retrieve a list of field types.
        """

        self.assertEqual(self.client.types(), TYPES_DATA["data"])
        self.assertEqual(self.config.token, TOKEN)

    @mock.patch("onyx.OnyxClient._request_handler", side_effect=mock_request)
    def test_lookups(self, mock_request):
        """
        Test that the OnyxClient can retrieve a list of lookups.
        """

        self.assertEqual(self.client.lookups(), LOOKUPS_DATA["data"])
        self.assertEqual(self.config.token, TOKEN)

    @mock.patch("onyx.OnyxClient._request_handler", side_effect=mock_request)
    def test_fields(self, mock_request):
        """
        Test that the OnyxClient can retrieve the field specification of a project.
        """

        self.assertEqual(self.client.fields(PROJECT), FIELDS_DATA["data"])
        self.assertEqual(self.config.token, TOKEN)

        for invalid in INVALID_ARGUMENTS:
            with pytest.raises(exceptions.OnyxClientError):
                self.client.fields(invalid)

    @mock.patch("onyx.OnyxClient._request_handler", side_effect=mock_request)
    def test_choices(self, mock_request):
        """
        Test that the OnyxClient can retrieve the choices of a choice field.
        """

        self.assertEqual(
            self.client.choices(PROJECT, CHOICE_FIELD), CHOICES_DATA["data"]
        )
        self.assertEqual(self.config.token, TOKEN)

        for invalid in INVALID_ARGUMENTS:
            with pytest.raises(exceptions.OnyxClientError):
                self.client.choices(invalid, CHOICE_FIELD)

            with pytest.raises(exceptions.OnyxClientError):
                self.client.choices(PROJECT, invalid)

    @mock.patch("onyx.OnyxClient._request_handler", side_effect=mock_request)
    def test_get(self, mock_request):
        """
        Test that the OnyxClient can retrieve a record from a project.
        """

        self.assertEqual(self.client.get(PROJECT, CLIMB_ID), GET_DATA["data"])
        self.assertEqual(
            self.client.get(
                PROJECT,
                fields={
                    "sample_id": SAMPLE_ID,
                    "run_name": RUN_NAME,
                    "published_date__range": ",".join(PUBLISHED_DATE_RANGE),
                },
            ),
            FILTER_SPECIFIC_DATA["data"][0],
        )
        self.assertEqual(
            self.client.get(
                PROJECT,
                fields={
                    "sample_id": SAMPLE_ID,
                    "run_name": RUN_NAME,
                    "published_date__range": ",".join(PUBLISHED_DATE_RANGE),
                },
                include=INCLUDE_FIELDS,
            ),
            FILTER_SPECIFIC_INCLUDE_DATA["data"][0],
        )
        self.assertEqual(
            self.client.get(
                PROJECT,
                fields={
                    "sample_id": SAMPLE_ID,
                    "run_name": RUN_NAME,
                    "published_date__range": ",".join(PUBLISHED_DATE_RANGE),
                },
                exclude=EXCLUDE_FIELDS,
            ),
            FILTER_SPECIFIC_EXCLUDE_DATA["data"][0],
        )
        self.assertEqual(self.config.token, TOKEN)

        # Cannot provide both CLIMB_ID and fields
        with pytest.raises(exceptions.OnyxClientError):
            self.client.get(
                PROJECT,
                CLIMB_ID,
                fields={"sample_id": SAMPLE_ID, "run_name": RUN_NAME},
            )

        # At least one of CLIMB_ID and fields is required
        with pytest.raises(exceptions.OnyxClientError):
            self.client.get(PROJECT)

        # More than one record returned
        with pytest.raises(exceptions.OnyxClientError):
            self.client.get(
                PROJECT, fields={"sample_id": "sample-123", "run_name": "run-456"}
            )

        for invalid in INVALID_ARGUMENTS:
            with pytest.raises(exceptions.OnyxClientError):
                self.client.get(invalid, CLIMB_ID)

            with pytest.raises(exceptions.OnyxClientError):
                self.client.get(PROJECT, invalid)

        for clash in PROJECT_ENDPOINT_CLASHES:
            with pytest.raises(exceptions.OnyxClientError):
                self.client.get(clash, CLIMB_ID)

        for clash in CLIMB_ID_ENDPOINT_CLASHES:
            with pytest.raises(exceptions.OnyxClientError):
                self.client.get(PROJECT, clash)

    @mock.patch("onyx.OnyxClient._request_handler", side_effect=mock_request)
    def test_filter(self, mock_request):
        """
        Test that the OnyxClient can filter records from a project.
        """

        self.assertEqual(
            [x for x in self.client.filter(PROJECT)],
            FILTER_PAGE_1_DATA["data"] + FILTER_PAGE_2_DATA["data"],
        )
        self.assertEqual(
            [
                x
                for x in self.client.filter(
                    PROJECT,
                    fields={
                        "sample_id": SAMPLE_ID,
                        "run_name": RUN_NAME,
                        "published_date__range": ",".join(PUBLISHED_DATE_RANGE),
                    },
                )
            ],
            FILTER_SPECIFIC_DATA["data"],
        )
        self.assertEqual(
            [
                x
                for x in self.client.filter(
                    PROJECT,
                    sample_id=SAMPLE_ID,
                    run_name=RUN_NAME,
                    published_date__range=PUBLISHED_DATE_RANGE,
                )
            ],
            FILTER_SPECIFIC_DATA["data"],
        )
        self.assertEqual(
            [
                x
                for x in self.client.filter(
                    PROJECT,
                    fields={"sample_id": "will-be-overwritten", "run_name": RUN_NAME},
                    sample_id=SAMPLE_ID,
                    published_date__range=PUBLISHED_DATE_RANGE,
                )
            ],
            FILTER_SPECIFIC_DATA["data"],
        )
        self.assertEqual(
            [
                x
                for x in self.client.filter(
                    PROJECT,
                    fields={
                        "sample_id": SAMPLE_ID,
                        "run_name": RUN_NAME,
                        "published_date__range": ",".join(PUBLISHED_DATE_RANGE),
                    },
                    include=INCLUDE_FIELDS,
                )
            ],
            FILTER_SPECIFIC_INCLUDE_DATA["data"],
        )
        self.assertEqual(
            [
                x
                for x in self.client.filter(
                    PROJECT,
                    fields={
                        "sample_id": SAMPLE_ID,
                        "run_name": RUN_NAME,
                        "published_date__range": ",".join(PUBLISHED_DATE_RANGE),
                    },
                    exclude=EXCLUDE_FIELDS,
                )
            ],
            FILTER_SPECIFIC_EXCLUDE_DATA["data"],
        )
        for empty in ["", None]:
            self.assertEqual(
                [
                    x
                    for x in self.client.filter(
                        PROJECT,
                        fields={
                            NONE_FIELD: empty,
                        },
                    )
                ],
                FILTER_NONE_DATA["data"],
            )
            self.assertEqual(
                [
                    x
                    for x in self.client.filter(
                        **{"project": PROJECT, NONE_FIELD: empty},
                    )
                ],
                FILTER_NONE_DATA["data"],
            )
            for type_ in [list, tuple, set]:
                self.assertEqual(
                    [
                        x
                        for x in self.client.filter(
                            PROJECT,
                            fields={
                                f"{NONE_FIELD}__in": type_([empty, "not-empty"]),
                            },
                        )
                    ],
                    FILTER_NONE_IN_DATA["data"],
                )
                self.assertEqual(
                    [
                        x
                        for x in self.client.filter(
                            **{
                                "project": PROJECT,
                                f"{NONE_FIELD}__in": type_([empty, "not-empty"]),
                            },
                        )
                    ],
                    FILTER_NONE_IN_DATA["data"],
                )
        self.assertEqual(self.config.token, TOKEN)

        for invalid in INVALID_ARGUMENTS:
            with pytest.raises(exceptions.OnyxClientError):
                [x for x in self.client.filter(invalid)]

    @mock.patch("onyx.OnyxClient._request_handler", side_effect=mock_request)
    def test_query(self, mock_request):
        """
        Test that the OnyxClient can query records from a project.
        """

        self.assertEqual(
            [x for x in self.client.query(PROJECT)],
            QUERY_PAGE_1_DATA["data"] + QUERY_PAGE_2_DATA["data"],
        )
        self.assertEqual(
            [
                x
                for x in self.client.query(
                    PROJECT,
                    query=OnyxField(sample_id=SAMPLE_ID) & OnyxField(run_name=RUN_NAME),
                )
            ],
            FILTER_SPECIFIC_DATA["data"],
        )
        self.assertEqual(
            [
                x
                for x in self.client.query(
                    PROJECT,
                    query=OnyxField(sample_id=SAMPLE_ID) & OnyxField(run_name=RUN_NAME),
                    include=INCLUDE_FIELDS,
                )
            ],
            FILTER_SPECIFIC_INCLUDE_DATA["data"],
        )
        self.assertEqual(
            [
                x
                for x in self.client.query(
                    PROJECT,
                    query=OnyxField(sample_id=SAMPLE_ID) & OnyxField(run_name=RUN_NAME),
                    exclude=EXCLUDE_FIELDS,
                )
            ],
            FILTER_SPECIFIC_EXCLUDE_DATA["data"],
        )
        self.assertEqual(self.config.token, TOKEN)

        with pytest.raises(exceptions.OnyxClientError):
            [
                x
                for x in self.client.query(
                    PROJECT, query="not_a_query_object"  #  type: ignore
                )
            ]

        for invalid in INVALID_ARGUMENTS:
            with pytest.raises(exceptions.OnyxClientError):
                [x for x in self.client.query(invalid)]

    @mock.patch("onyx.OnyxClient._request_handler", side_effect=mock_request)
    def test_to_csv(self, mock_request):
        """
        Test that the OnyxClient can convert records to CSV.
        """

        pass  # TODO Test to_csv

    @mock.patch("onyx.OnyxClient._request_handler", side_effect=mock_request)
    def test_history(self, mock_request):
        """
        Test that the OnyxClient can retrieve the history of a record.
        """

        self.assertEqual(self.client.history(PROJECT, CLIMB_ID), HISTORY_DATA["data"])
        self.assertEqual(self.config.token, TOKEN)

        for invalid in INVALID_ARGUMENTS:
            with pytest.raises(exceptions.OnyxClientError):
                self.client.history(invalid, CLIMB_ID)

            with pytest.raises(exceptions.OnyxClientError):
                self.client.history(PROJECT, invalid)

        for clash in PROJECT_ENDPOINT_CLASHES:
            with pytest.raises(exceptions.OnyxClientError):
                self.client.get(clash, CLIMB_ID)

        for clash in CLIMB_ID_ENDPOINT_CLASHES:
            with pytest.raises(exceptions.OnyxClientError):
                self.client.history(PROJECT, clash)

    @mock.patch("onyx.OnyxClient._request_handler", side_effect=mock_request)
    def test_identify(self, mock_request):
        """
        Test that the OnyxClient can identify an anonymised value on a field.
        """

        self.assertEqual(
            self.client.identify(PROJECT, IDENTIFY_FIELD, IDENTIFY_VALUE),
            IDENTIFY_DATA["data"],
        )
        self.assertEqual(
            self.client.identify(
                PROJECT, IDENTIFY_FIELD, IDENTIFY_VALUE, site=OTHER_SITE
            ),
            IDENTIFY_OTHER_SITE_DATA["data"],
        )
        self.assertEqual(self.config.token, TOKEN)

        for invalid in INVALID_ARGUMENTS:
            with pytest.raises(exceptions.OnyxClientError):
                self.client.identify(invalid, IDENTIFY_FIELD, IDENTIFY_VALUE)

            with pytest.raises(exceptions.OnyxClientError):
                self.client.identify(
                    invalid, IDENTIFY_FIELD, IDENTIFY_VALUE, site=OTHER_SITE
                )

            with pytest.raises(exceptions.OnyxClientError):
                self.client.identify(PROJECT, invalid, IDENTIFY_VALUE)

            with pytest.raises(exceptions.OnyxClientError):
                self.client.identify(PROJECT, invalid, IDENTIFY_VALUE, site=OTHER_SITE)

    @mock.patch("onyx.OnyxClient._request_handler", side_effect=mock_request)
    def test_create(self, mock_request):
        """
        Test that the OnyxClient can create a record in a project.
        """

        self.assertEqual(
            self.client.create(PROJECT, CREATE_FIELDS), CREATE_DATA["data"]
        )
        self.assertEqual(
            self.client.create(PROJECT, CREATE_FIELDS, test=True),
            TESTCREATE_DATA["data"],
        )
        self.assertEqual(self.config.token, TOKEN)

        for invalid in INVALID_ARGUMENTS:
            with pytest.raises(exceptions.OnyxClientError):
                self.client.create(invalid, CREATE_FIELDS)

            with pytest.raises(exceptions.OnyxClientError):
                self.client.create(invalid, CREATE_FIELDS, test=True)

    @mock.patch("onyx.OnyxClient._request_handler", side_effect=mock_request)
    def test_update(self, mock_request):
        """
        Test that the OnyxClient can update a record in a project.
        """

        self.assertEqual(
            self.client.update(PROJECT, CLIMB_ID, UPDATE_FIELDS), UPDATE_DATA["data"]
        )
        self.assertEqual(
            self.client.update(PROJECT, CLIMB_ID, UPDATE_FIELDS, test=True),
            TESTUPDATE_DATA["data"],
        )
        self.assertEqual(self.config.token, TOKEN)

        for invalid in INVALID_ARGUMENTS:
            with pytest.raises(exceptions.OnyxClientError):
                self.client.update(invalid, CLIMB_ID, UPDATE_FIELDS)

            with pytest.raises(exceptions.OnyxClientError):
                self.client.update(invalid, CLIMB_ID, UPDATE_FIELDS, test=True)

            with pytest.raises(exceptions.OnyxClientError):
                self.client.update(PROJECT, invalid, UPDATE_FIELDS)

            with pytest.raises(exceptions.OnyxClientError):
                self.client.update(PROJECT, invalid, UPDATE_FIELDS, test=True)

        for clash in PROJECT_ENDPOINT_CLASHES:
            with pytest.raises(exceptions.OnyxClientError):
                self.client.get(clash, CLIMB_ID, UPDATE_FIELDS)

        for clash in CLIMB_ID_ENDPOINT_CLASHES:
            with pytest.raises(exceptions.OnyxClientError):
                self.client.update(PROJECT, clash, UPDATE_FIELDS)

    @mock.patch("onyx.OnyxClient._request_handler", side_effect=mock_request)
    def test_delete(self, mock_request):
        """
        Test that the OnyxClient can delete a record from a project.
        """

        self.assertEqual(self.client.delete(PROJECT, CLIMB_ID), DELETE_DATA["data"])
        self.assertEqual(self.config.token, TOKEN)

        for invalid in INVALID_ARGUMENTS:
            with pytest.raises(exceptions.OnyxClientError):
                self.client.delete(invalid, CLIMB_ID)

            with pytest.raises(exceptions.OnyxClientError):
                self.client.delete(PROJECT, invalid)

    @mock.patch("onyx.OnyxClient._request_handler", side_effect=mock_request)
    def test_csv_create(self, mock_request):
        """
        Test that the OnyxClient can create records from a CSV file.
        """

        for data, test in [
            (CREATE_DATA["data"], False),
            (TESTCREATE_DATA["data"], True),
        ]:
            self.assertEqual(
                self.client.csv_create(
                    PROJECT,
                    io.StringIO(CSV_CREATE_SINGLE_FILE),
                    test=test,
                ),
                data,
            )
            self.assertEqual(
                self.client.csv_create(
                    PROJECT,
                    io.StringIO(TSV_CREATE_SINGLE_FILE),
                    delimiter="\t",
                    test=test,
                ),
                data,
            )
            self.assertEqual(
                self.client.csv_create(
                    PROJECT,
                    io.StringIO(CSV_CREATE_MULTI_FILE),
                    multiline=True,
                    test=test,
                ),
                [data, data],
            )
            self.assertEqual(
                self.client.csv_create(
                    PROJECT,
                    io.StringIO(TSV_CREATE_MULTI_FILE),
                    delimiter="\t",
                    multiline=True,
                    test=test,
                ),
                [data, data],
            )
            self.assertEqual(
                self.client.csv_create(
                    PROJECT,
                    io.StringIO(CSV_CREATE_SINGLE_MISSING_FILE),
                    fields=MISSING_CREATE_FIELDS,
                    test=test,
                ),
                data,
            )
            self.assertEqual(
                self.client.csv_create(
                    PROJECT,
                    io.StringIO(TSV_CREATE_SINGLE_MISSING_FILE),
                    fields=MISSING_CREATE_FIELDS,
                    delimiter="\t",
                    test=test,
                ),
                data,
            )
            self.assertEqual(
                self.client.csv_create(
                    PROJECT,
                    io.StringIO(CSV_CREATE_MULTI_MISSING_FILE),
                    fields=MISSING_CREATE_FIELDS,
                    multiline=True,
                    test=test,
                ),
                [data, data],
            )
            self.assertEqual(
                self.client.csv_create(
                    PROJECT,
                    io.StringIO(TSV_CREATE_MULTI_MISSING_FILE),
                    fields=MISSING_CREATE_FIELDS,
                    delimiter="\t",
                    multiline=True,
                    test=test,
                ),
                [data, data],
            )
        self.assertEqual(self.config.token, TOKEN)

        with pytest.raises(exceptions.OnyxClientError):
            self.client.csv_create(
                PROJECT,
                io.StringIO(CSV_CREATE_EMPTY_FILE),
            )

        with pytest.raises(exceptions.OnyxClientError):
            self.client.csv_create(
                PROJECT,
                io.StringIO(TSV_CREATE_EMPTY_FILE),
                delimiter="\t",
            )

        with pytest.raises(exceptions.OnyxClientError):
            self.client.csv_create(
                PROJECT,
                io.StringIO(CSV_CREATE_MULTI_FILE),
            )

        with pytest.raises(exceptions.OnyxClientError):
            self.client.csv_create(
                PROJECT,
                io.StringIO(TSV_CREATE_MULTI_FILE),
                delimiter="\t",
            )

    @mock.patch("onyx.OnyxClient._request_handler", side_effect=mock_request)
    def test_csv_update(self, mock_request):
        """
        Test that the OnyxClient can update records from a CSV file.
        """

        for data, test in [
            (UPDATE_DATA["data"], False),
            (TESTUPDATE_DATA["data"], True),
        ]:
            self.assertEqual(
                self.client.csv_update(
                    PROJECT,
                    io.StringIO(CSV_UPDATE_SINGLE_FILE),
                    test=test,
                ),
                data,
            )
            self.assertEqual(
                self.client.csv_update(
                    PROJECT,
                    io.StringIO(TSV_UPDATE_SINGLE_FILE),
                    delimiter="\t",
                    test=test,
                ),
                data,
            )
            self.assertEqual(
                self.client.csv_update(
                    PROJECT,
                    io.StringIO(CSV_UPDATE_MULTI_FILE),
                    multiline=True,
                    test=test,
                ),
                [data, data],
            )
            self.assertEqual(
                self.client.csv_update(
                    PROJECT,
                    io.StringIO(TSV_UPDATE_MULTI_FILE),
                    delimiter="\t",
                    multiline=True,
                    test=test,
                ),
                [data, data],
            )
            self.assertEqual(
                self.client.csv_update(
                    PROJECT,
                    io.StringIO(CSV_UPDATE_SINGLE_MISSING_FILE),
                    fields=MISSING_UPDATE_FIELDS,
                    test=test,
                ),
                data,
            )
            self.assertEqual(
                self.client.csv_update(
                    PROJECT,
                    io.StringIO(TSV_UPDATE_SINGLE_MISSING_FILE),
                    fields=MISSING_UPDATE_FIELDS,
                    delimiter="\t",
                    test=test,
                ),
                data,
            )
            self.assertEqual(
                self.client.csv_update(
                    PROJECT,
                    io.StringIO(CSV_UPDATE_MULTI_MISSING_FILE),
                    fields=MISSING_UPDATE_FIELDS,
                    multiline=True,
                    test=test,
                ),
                [data, data],
            )
            self.assertEqual(
                self.client.csv_update(
                    PROJECT,
                    io.StringIO(TSV_UPDATE_MULTI_MISSING_FILE),
                    fields=MISSING_UPDATE_FIELDS,
                    delimiter="\t",
                    multiline=True,
                    test=test,
                ),
                [data, data],
            )
        self.assertEqual(self.config.token, TOKEN)

        with pytest.raises(exceptions.OnyxClientError):
            self.client.csv_update(
                PROJECT,
                io.StringIO(CSV_UPDATE_EMPTY_FILE),
            )

        with pytest.raises(exceptions.OnyxClientError):
            self.client.csv_update(
                PROJECT,
                io.StringIO(TSV_UPDATE_EMPTY_FILE),
                delimiter="\t",
            )

        with pytest.raises(exceptions.OnyxClientError):
            self.client.csv_update(
                PROJECT,
                io.StringIO(CSV_UPDATE_MULTI_FILE),
            )

        with pytest.raises(exceptions.OnyxClientError):
            self.client.csv_update(
                PROJECT,
                io.StringIO(TSV_UPDATE_MULTI_FILE),
                delimiter="\t",
            )

        # Testing lack of CLIMB_ID for update
        with pytest.raises(exceptions.OnyxClientError):
            self.client.csv_update(
                PROJECT,
                io.StringIO(CSV_UPDATE_SINGLE_MISSING_FILE),
            )

        with pytest.raises(exceptions.OnyxClientError):
            self.client.csv_update(
                PROJECT,
                io.StringIO(CSV_UPDATE_MULTI_MISSING_FILE),
                multiline=True,
            )

        with pytest.raises(exceptions.OnyxClientError):
            self.client.csv_update(
                PROJECT,
                io.StringIO(TSV_UPDATE_SINGLE_MISSING_FILE),
                delimiter="\t",
            )

        with pytest.raises(exceptions.OnyxClientError):
            self.client.csv_update(
                PROJECT,
                io.StringIO(TSV_UPDATE_MULTI_MISSING_FILE),
                delimiter="\t",
                multiline=True,
            )

    @mock.patch("onyx.OnyxClient._request_handler", side_effect=mock_request)
    def test_csv_delete(self, mock_request):
        """
        Test that the OnyxClient can delete records from a CSV file.
        """

        self.assertEqual(
            self.client.csv_delete(
                PROJECT,
                io.StringIO(CSV_DELETE_SINGLE_FILE),
            ),
            DELETE_DATA["data"],
        )
        self.assertEqual(
            self.client.csv_delete(
                PROJECT,
                io.StringIO(TSV_DELETE_SINGLE_FILE),
                delimiter="\t",
            ),
            DELETE_DATA["data"],
        )
        self.assertEqual(
            self.client.csv_delete(
                PROJECT,
                io.StringIO(CSV_DELETE_MULTI_FILE),
                multiline=True,
            ),
            [DELETE_DATA["data"], DELETE_DATA["data"]],
        )
        self.assertEqual(
            self.client.csv_delete(
                PROJECT,
                io.StringIO(TSV_DELETE_MULTI_FILE),
                delimiter="\t",
                multiline=True,
            ),
            [DELETE_DATA["data"], DELETE_DATA["data"]],
        )
        self.assertEqual(self.config.token, TOKEN)

        with pytest.raises(exceptions.OnyxClientError):
            self.client.csv_delete(
                PROJECT,
                io.StringIO(CSV_DELETE_EMPTY_FILE),
            )

        with pytest.raises(exceptions.OnyxClientError):
            self.client.csv_delete(
                PROJECT,
                io.StringIO(TSV_DELETE_EMPTY_FILE),
                delimiter="\t",
            )

        with pytest.raises(exceptions.OnyxClientError):
            self.client.csv_delete(
                PROJECT,
                io.StringIO(CSV_DELETE_MULTI_FILE),
            )

        with pytest.raises(exceptions.OnyxClientError):
            self.client.csv_delete(
                PROJECT,
                io.StringIO(TSV_DELETE_MULTI_FILE),
                delimiter="\t",
            )

    @mock.patch("requests.post", side_effect=mock_register_post)
    def test_register(self, mock_request):
        """
        Test that the OnyxClient can register a new user.
        """

        self.assertEqual(
            OnyxClient.register(
                domain=DOMAIN,
                first_name=FIRST_NAME,
                last_name=LAST_NAME,
                email=EMAIL,
                site=SITE,
                password=PASSWORD,
            ),
            REGISTER_DATA["data"],
        )

        for invalid in INVALID_DOMAIN_ARGUMENTS:
            with pytest.raises(exceptions.OnyxClientError):
                OnyxClient.register(
                    domain=invalid,
                    first_name=FIRST_NAME,
                    last_name=LAST_NAME,
                    email=EMAIL,
                    site=SITE,
                    password=PASSWORD,
                )

    @mock.patch("onyx.OnyxClient._request_handler", side_effect=mock_request)
    def test_login(self, mock_request):
        """
        Test that the OnyxClient can login.
        """

        self.assertEqual(self.client.login(), LOGIN_DATA["data"])
        self.assertEqual(self.config.token, TOKEN)

    @mock.patch("onyx.OnyxClient._request_handler", side_effect=mock_request)
    def test_logout(self, mock_request):
        """
        Test that the OnyxClient can logout.
        """

        self.assertEqual(self.client.logout(), LOGOUT_DATA["data"])
        self.assertEqual(self.config.token, None)

    @mock.patch("onyx.OnyxClient._request_handler", side_effect=mock_request)
    def test_logoutall(self, mock_request):
        """
        Test that the OnyxClient can logout from all devices.
        """

        self.assertEqual(self.client.logoutall(), LOGOUTALL_DATA["data"])
        self.assertEqual(self.config.token, None)

    @mock.patch("onyx.OnyxClient._request_handler", side_effect=mock_request)
    def test_profile(self, mock_request):
        """
        Test that the OnyxClient can retrieve the user profile.
        """

        self.assertEqual(self.client.profile(), PROFILE_DATA["data"])
        self.assertEqual(self.config.token, TOKEN)

    @mock.patch("onyx.OnyxClient._request_handler", side_effect=mock_request)
    def test_activity(self, mock_request):
        """
        Test that the OnyxClient can retrieve the user's latest activity.
        """

        self.assertEqual(self.client.activity(), ACTIVITY_DATA["data"])
        self.assertEqual(self.config.token, TOKEN)

    @mock.patch("onyx.OnyxClient._request_handler", side_effect=mock_request)
    def test_approve(self, mock_request):
        """
        Test that the OnyxClient can approve a user.
        """

        self.assertEqual(self.client.approve(OTHER_USERNAME), APPROVE_DATA["data"])
        self.assertEqual(self.config.token, TOKEN)

        for invalid in INVALID_ARGUMENTS:
            with pytest.raises(exceptions.OnyxClientError):
                self.client.approve(invalid)

    @mock.patch("onyx.OnyxClient._request_handler", side_effect=mock_request)
    def test_waiting(self, mock_request):
        """
        Test that the OnyxClient can retrieve a list of users waiting to be approved.
        """

        self.assertEqual(self.client.waiting(), WAITING_DATA["data"])
        self.assertEqual(self.config.token, TOKEN)

    @mock.patch("onyx.OnyxClient._request_handler", side_effect=mock_request)
    def test_site_users(self, mock_request):
        """
        Test that the OnyxClient can retrieve a list of users on the site.
        """

        self.assertEqual(self.client.site_users(), SITE_USERS_DATA["data"])
        self.assertEqual(self.config.token, TOKEN)

    @mock.patch("onyx.OnyxClient._request_handler", side_effect=mock_request)
    def test_all_users(self, mock_request):
        """
        Test that the OnyxClient can retrieve a list of all users.
        """

        self.assertEqual(self.client.all_users(), ALL_USERS_DATA["data"])
        self.assertEqual(self.config.token, TOKEN)



import posixpath
import csv
import inspect
import requests
from requests import HTTPError, RequestException
from typing import Any, Generator, List, Dict, TextIO, Optional, Union
from .config import OnyxConfig
from .field import OnyxField
from .exceptions import (
    OnyxClientError,
    OnyxConnectionError,
    OnyxRequestError,
    OnyxServerError,
)


class OnyxClientBase:
    __slots__ = "config", "_request_handler", "_session"
    ENDPOINTS = {
        "projects": lambda domain: OnyxClient._handle_endpoint(
            lambda: posixpath.join(
                str(domain),
                "projects/",
            ),
            domain=domain,
        ),
        "types": lambda domain: OnyxClient._handle_endpoint(
            lambda: posixpath.join(
                str(domain),
                "projects/types/",
            ),
            domain=domain,
        ),
        "lookups": lambda domain: OnyxClient._handle_endpoint(
            lambda: posixpath.join(
                str(domain),
                "projects/lookups/",
            ),
            domain=domain,
        ),
        "fields": lambda domain, project: OnyxClient._handle_endpoint(
            lambda: posixpath.join(
                str(domain),
                "projects",
                str(project),
                "fields/",
            ),
            domain=domain,
            project=project,
        ),
        "choices": lambda domain, project, field: OnyxClient._handle_endpoint(
            lambda: posixpath.join(
                str(domain),
                "projects",
                str(project),
                "choices",
                str(field),
                "",
            ),
            domain=domain,
            project=project,
            field=field,
        ),
        "get": lambda domain, project, climb_id: OnyxClient._handle_endpoint(
            lambda: posixpath.join(
                str(domain),
                "projects",
                str(project),
                str(climb_id),
                "",
            ),
            domain=domain,
            project=project,
            climb_id=climb_id,
        ),
        "filter": lambda domain, project: OnyxClient._handle_endpoint(
            lambda: posixpath.join(
                str(domain),
                "projects",
                str(project),
                "",
            ),
            domain=domain,
            project=project,
        ),
        "query": lambda domain, project: OnyxClient._handle_endpoint(
            lambda: posixpath.join(
                str(domain),
                "projects",
                str(project),
                "query/",
            ),
            domain=domain,
            project=project,
        ),
        "history": lambda domain, project, climb_id: OnyxClient._handle_endpoint(
            lambda: posixpath.join(
                str(domain),
                "projects",
                str(project),
                "history",
                str(climb_id),
                "",
            ),
            domain=domain,
            project=project,
            climb_id=climb_id,
        ),
        "identify": lambda domain, project, field: OnyxClient._handle_endpoint(
            lambda: posixpath.join(
                str(domain),
                "projects",
                str(project),
                "identify",
                str(field),
                "",
            ),
            domain=domain,
            project=project,
            field=field,
        ),
        "create": lambda domain, project: OnyxClient._handle_endpoint(
            lambda: posixpath.join(
                str(domain),
                "projects",
                str(project),
                "",
            ),
            domain=domain,
            project=project,
        ),
        "testcreate": lambda domain, project: OnyxClient._handle_endpoint(
            lambda: posixpath.join(
                str(domain),
                "projects",
                str(project),
                "test/",
            ),
            domain=domain,
            project=project,
        ),
        "update": lambda domain, project, climb_id: OnyxClient._handle_endpoint(
            lambda: posixpath.join(
                str(domain),
                "projects",
                str(project),
                str(climb_id),
                "",
            ),
            domain=domain,
            project=project,
            climb_id=climb_id,
        ),
        "testupdate": lambda domain, project, climb_id: OnyxClient._handle_endpoint(
            lambda: posixpath.join(
                str(domain),
                "projects",
                str(project),
                "test",
                str(climb_id),
                "",
            ),
            domain=domain,
            project=project,
            climb_id=climb_id,
        ),
        "delete": lambda domain, project, climb_id: OnyxClient._handle_endpoint(
            lambda: posixpath.join(
                str(domain),
                "projects",
                str(project),
                str(climb_id),
                "",
            ),
            domain=domain,
            project=project,
            climb_id=climb_id,
        ),
        "register": lambda domain: OnyxClient._handle_endpoint(
            lambda: posixpath.join(
                str(domain),
                "accounts/register/",
            ),
            domain=domain,
        ),
        "login": lambda domain: OnyxClient._handle_endpoint(
            lambda: posixpath.join(
                str(domain),
                "accounts/login/",
            ),
            domain=domain,
        ),
        "logout": lambda domain: OnyxClient._handle_endpoint(
            lambda: posixpath.join(
                str(domain),
                "accounts/logout/",
            ),
            domain=domain,
        ),
        "logoutall": lambda domain: OnyxClient._handle_endpoint(
            lambda: posixpath.join(
                str(domain),
                "accounts/logoutall/",
            ),
            domain=domain,
        ),
        "profile": lambda domain: OnyxClient._handle_endpoint(
            lambda: posixpath.join(
                str(domain),
                "accounts/profile/",
            ),
            domain=domain,
        ),
        "activity": lambda domain: OnyxClient._handle_endpoint(
            lambda: posixpath.join(
                str(domain),
                "accounts/activity/",
            ),
            domain=domain,
        ),
        "waiting": lambda domain: OnyxClient._handle_endpoint(
            lambda: posixpath.join(
                str(domain),
                "accounts/waiting/",
            ),
            domain=domain,
        ),
        "approve": lambda domain, username: OnyxClient._handle_endpoint(
            lambda: posixpath.join(
                str(domain),
                "accounts/approve",
                str(username),
                "",
            ),
            domain=domain,
            username=username,
        ),
        "siteusers": lambda domain: OnyxClient._handle_endpoint(
            lambda: posixpath.join(
                str(domain),
                "accounts/site/",
            ),
            domain=domain,
        ),
        "allusers": lambda domain: OnyxClient._handle_endpoint(
            lambda: posixpath.join(
                str(domain),
                "accounts/all/",
            ),
            domain=domain,
        ),
    }

    def __init__(self, config: OnyxConfig):
        self.config = config
        self._session = None
        self._request_handler = requests.request

    def __enter__(self):
        self._session = requests.Session()
        self._request_handler = self._session.request
        return self

    def __exit__(self, type, value, traceback):
        if self._session:
            self._session.close()
        self._request_handler = requests.request

    @classmethod
    def _handle_endpoint(cls, endpoint, **kwargs):
        for name, val in kwargs.items():
            if val is None or not str(val).strip():
                raise OnyxClientError(f"Argument '{name}' was not provided.")

            val = str(val).strip()

            if name != "domain":
                for char in "/?":
                    if char in val:
                        raise OnyxClientError(
                            f"Argument '{name}' contains invalid character: '{char}'."
                        )

                # Crude but effective prevention of unexpectedly calling other endpoints
                # Its not the end of the world if that did happen, but to the user it would be quite confusing
                clashes = {
                    "project": ["types", "lookups"],
                    "climb_id": [
                        "test",
                        "query",
                        "fields",
                        "choices",
                        "history",
                        "identify",
                    ],
                }

                if name in clashes:
                    for clash in clashes[name]:
                        if val == clash:
                            raise OnyxClientError(
                                f"Argument '{name}' cannot have value '{val}'. This creates a URL that resolves to a different endpoint."
                            )

        return endpoint()

    def _request(self, method: str, retries: int = 3, **kwargs) -> requests.Response:
        if not retries:
            raise Exception(
                "Request retry limit reached. This should not be possible..."
            )

        kwargs.setdefault("headers", {}).update(
            {"Authorization": f"Token {self.config.token}"}
        )
        method_response = self._request_handler(method, **kwargs)

        # Token has expired or was invalid.
        # If username and password were provided, log in again, obtain a new token, and re-run the method.
        if (
            method_response.status_code == 401
            and self.config.username
            and self.config.password
        ):
            OnyxClientBase.login(self).raise_for_status()
            # A retry mechanism has been incorporated as a failsafe.
            # This is to protect against the case where an onyx endpoint returns a 401 status code,
            # despite the user being able to successfully log in, leading to an infinite loop of
            # re-logging in and re-hitting the endpoint.
            # This scenario should not be possible. But if it happened, it would not be fun at all.
            # So, better safe than sorry...
            return self._request(method, retries=retries - 1, **kwargs)

        return method_response

    def _csv_upload(
        self,
        method: str,
        endpoint: str,
        project: str,
        csv_file: TextIO,
        fields: Optional[Dict[str, Any]] = None,
        delimiter: Optional[str] = None,
        multiline: bool = False,
        test: bool = False,
        climb_id_required: bool = False,
    ) -> Generator[requests.Response, Any, None]:
        # Get appropriate endpoint for test/prod
        if test:
            endpoint = "test" + endpoint

        # Create CSV reader
        if delimiter is None:
            reader = csv.DictReader(
                csv_file,
                skipinitialspace=True,
            )
        else:
            reader = csv.DictReader(
                csv_file,
                delimiter=delimiter,
                skipinitialspace=True,
            )

        # Read the first two records (if they exist) and store in 'records' list
        # This is done to protect against two scenarios:
        # - There are no records in the file (never allowed)
        # - There is more than one record, but multiline = False (not allowed)
        records = []

        record_1 = next(reader, None)
        if record_1:
            records.append(record_1)
        else:
            raise OnyxClientError("File must contain at least one record.")

        record_2 = next(reader, None)
        if record_2:
            if not multiline:
                raise OnyxClientError(
                    "File contains multiple records but this is not allowed. To upload multiple records, set 'multiline' = True."
                )
            records.append(record_2)

        # Iterate over the read and unread records and upload sequentially
        for iterator in (records, reader):
            for record in iterator:
                if fields:
                    record = record | fields

                if climb_id_required:
                    # Grab the climb_id, if required for the URL
                    climb_id = record.pop("climb_id", None)
                    if not climb_id:
                        raise OnyxClientError(
                            "Record requires a 'climb_id' for upload."
                        )
                    url = OnyxClient.ENDPOINTS[endpoint](
                        self.config.domain, project, climb_id
                    )
                else:
                    url = OnyxClient.ENDPOINTS[endpoint](self.config.domain, project)

                response = self._request(
                    method=method,
                    url=url,
                    json=record,
                )
                yield response

    def _csv_handle_multiline(
        self,
        responses: Generator[requests.Response, Any, None],
        multiline: bool,
    ) -> Union[Dict[str, Any], List[Dict[str, Any]]]:
        if multiline:
            results = []
            for response in responses:
                response.raise_for_status()
                results.append(response.json()["data"])
            return results
        else:
            response = next(responses, None)
            if response is None:
                raise OnyxClientError("Iterator must contain at least one record.")

            response.raise_for_status()
            return response.json()["data"]

    def projects(self) -> requests.Response:
        response = self._request(
            method="get",
            url=OnyxClient.ENDPOINTS["projects"](self.config.domain),
        )
        return response

    def types(self) -> requests.Response:
        response = self._request(
            method="get",
            url=OnyxClient.ENDPOINTS["types"](self.config.domain),
        )
        return response

    def lookups(self) -> requests.Response:
        response = self._request(
            method="get",
            url=OnyxClient.ENDPOINTS["lookups"](self.config.domain),
        )
        return response

    def fields(self, project: str) -> requests.Response:
        response = self._request(
            method="get",
            url=OnyxClient.ENDPOINTS["fields"](self.config.domain, project),
        )
        return response

    def choices(self, project: str, field: str) -> requests.Response:
        response = self._request(
            method="get",
            url=OnyxClient.ENDPOINTS["choices"](self.config.domain, project, field),
        )
        return response

    def get(
        self,
        project: str,
        climb_id: str,
        include: Union[List[str], str, None] = None,
        exclude: Union[List[str], str, None] = None,
    ) -> requests.Response:
        response = self._request(
            method="get",
            url=OnyxClient.ENDPOINTS["get"](self.config.domain, project, climb_id),
            params={"include": include, "exclude": exclude},
        )
        return response

    def filter(
        self,
        project: str,
        fields: Optional[Dict[str, Any]] = None,
        include: Union[List[str], str, None] = None,
        exclude: Union[List[str], str, None] = None,
        summarise: Union[List[str], str, None] = None,
        **kwargs: Any,
    ) -> Generator[requests.Response, Any, None]:
        if fields is None:
            fields = {}

        for field, value in kwargs.items():
            if type(value) in {list, tuple, set}:
                value = ",".join(map(lambda x: str(x) if x is not None else "", value))
            fields[field] = value

        for field, value in fields.items():
            if type(value) in {list, tuple, set}:
                fields[field] = [v if v is not None else "" for v in value]
            if value is None:
                fields[field] = ""

        fields = fields | {
            "include": include,
            "exclude": exclude,
            "summarise": summarise,
        }

        _next = OnyxClient.ENDPOINTS["filter"](self.config.domain, project)

        while _next is not None:
            response = self._request(
                method="get",
                url=_next,
                params=fields,
            )
            yield response

            fields = None
            if response.ok:
                _next = response.json().get("next")
            else:
                _next = None

    def query(
        self,
        project: str,
        query: Optional[OnyxField] = None,
        include: Union[List[str], str, None] = None,
        exclude: Union[List[str], str, None] = None,
        summarise: Union[List[str], str, None] = None,
    ) -> Generator[requests.Response, Any, None]:
        if query:
            if not isinstance(query, OnyxField):
                raise OnyxClientError(
                    f"Query must be an instance of {OnyxField}. Received: {type(query)}"
                )
            else:
                query_json = query.query
        else:
            query_json = None

        fields = {
            "include": include,
            "exclude": exclude,
            "summarise": summarise,
        }
        _next = OnyxClient.ENDPOINTS["query"](self.config.domain, project)

        while _next is not None:
            response = self._request(
                method="post",
                url=_next,
                json=query_json,
                params=fields,
            )
            yield response

            fields = None
            if response.ok:
                _next = response.json().get("next")
            else:
                _next = None

    @classmethod
    def to_csv(
        cls,
        csv_file: TextIO,
        data: Union[List[Dict[str, Any]], Generator[Dict[str, Any], Any, None]],
        delimiter: Optional[str] = None,
    ):
        # Ensure data is an iterator
        if inspect.isgenerator(data):
            data_iterator = data
        else:
            data_iterator = iter(data)

        row = next(data_iterator, None)
        if row:
            fields = row.keys()

            # Create CSV writer
            if delimiter is None:
                writer = csv.DictWriter(
                    csv_file,
                    fieldnames=fields,
                )
            else:
                writer = csv.DictWriter(
                    csv_file,
                    fieldnames=fields,
                    delimiter=delimiter,
                )

            # Write data
            writer.writeheader()
            writer.writerow(row)
            for row in data_iterator:
                writer.writerow(row)

    def history(
        self,
        project: str,
        climb_id: str,
    ) -> requests.Response:
        response = self._request(
            method="get",
            url=OnyxClient.ENDPOINTS["history"](self.config.domain, project, climb_id),
        )
        return response

    def identify(
        self,
        project: str,
        field: str,
        value: str,
        site: Optional[str] = None,
    ) -> requests.Response:
        identify_json = {"value": value}
        if site:
            identify_json = identify_json | {"site": site}

        response = self._request(
            method="post",
            url=OnyxClient.ENDPOINTS["identify"](self.config.domain, project, field),
            json=identify_json,
        )
        return response

    def create(
        self,
        project: str,
        fields: Dict[str, Any],
        test: bool = False,
    ) -> requests.Response:
        if test:
            endpoint = "testcreate"
        else:
            endpoint = "create"

        response = self._request(
            method="post",
            url=OnyxClient.ENDPOINTS[endpoint](self.config.domain, project),
            json=fields,
        )
        return response

    def update(
        self,
        project: str,
        climb_id: str,
        fields: Optional[Dict[str, Any]] = None,
        test: bool = False,
    ) -> requests.Response:
        if test:
            endpoint = "testupdate"
        else:
            endpoint = "update"

        response = self._request(
            method="patch",
            url=OnyxClient.ENDPOINTS[endpoint](self.config.domain, project, climb_id),
            json=fields,
        )
        return response

    def delete(
        self,
        project: str,
        climb_id: str,
    ) -> requests.Response:
        response = self._request(
            method="delete",
            url=OnyxClient.ENDPOINTS["delete"](self.config.domain, project, climb_id),
        )
        return response

    def csv_create(
        self,
        project: str,
        csv_file: TextIO,
        fields: Optional[Dict[str, Any]] = None,
        delimiter: Optional[str] = None,
        multiline: bool = False,
        test: bool = False,
    ) -> Generator[requests.Response, Any, None]:
        yield from self._csv_upload(
            method="post",
            endpoint="create",
            project=project,
            csv_file=csv_file,
            fields=fields,
            delimiter=delimiter,
            multiline=multiline,
            test=test,
        )

    def csv_update(
        self,
        project: str,
        csv_file: TextIO,
        fields: Optional[Dict[str, Any]] = None,
        delimiter: Optional[str] = None,
        multiline: bool = False,
        test: bool = False,
    ) -> Generator[requests.Response, Any, None]:
        yield from self._csv_upload(
            method="patch",
            endpoint="update",
            project=project,
            csv_file=csv_file,
            fields=fields,
            delimiter=delimiter,
            multiline=multiline,
            test=test,
            climb_id_required=True,
        )

    def csv_delete(
        self,
        project: str,
        csv_file: TextIO,
        delimiter: Optional[str] = None,
        multiline: bool = False,
    ) -> Generator[requests.Response, Any, None]:
        yield from self._csv_upload(
            method="delete",
            endpoint="delete",
            project=project,
            csv_file=csv_file,
            delimiter=delimiter,
            multiline=multiline,
            climb_id_required=True,
        )

    @classmethod
    def register(
        cls,
        domain: str,
        first_name: str,
        last_name: str,
        email: str,
        site: str,
        password: str,
    ) -> requests.Response:
        response = requests.post(
            OnyxClient.ENDPOINTS["register"](domain),
            json={
                "first_name": first_name,
                "last_name": last_name,
                "password": password,
                "email": email,
                "site": site,
            },
        )
        return response

    def login(self) -> requests.Response:
        if self.config.username and self.config.password:
            credentials = (self.config.username, self.config.password)
        else:
            credentials = None

        response = self._request_handler(
            "post",
            auth=credentials,
            url=OnyxClient.ENDPOINTS["login"](self.config.domain),
        )
        if response.ok:
            self.config.token = response.json()["data"]["token"]

        return response

    def logout(self) -> requests.Response:
        response = self._request(
            method="post",
            url=OnyxClient.ENDPOINTS["logout"](self.config.domain),
        )
        if response.ok:
            self.config.token = None

        return response

    def logoutall(self) -> requests.Response:
        response = self._request(
            method="post",
            url=OnyxClient.ENDPOINTS["logoutall"](self.config.domain),
        )
        if response.ok:
            self.config.token = None

        return response

    def profile(self) -> requests.Response:
        response = self._request(
            method="get",
            url=OnyxClient.ENDPOINTS["profile"](self.config.domain),
        )
        return response

    def activity(self) -> requests.Response:
        response = self._request(
            method="get",
            url=OnyxClient.ENDPOINTS["activity"](self.config.domain),
        )
        return response

    def approve(self, username: str) -> requests.Response:
        response = self._request(
            method="patch",
            url=OnyxClient.ENDPOINTS["approve"](self.config.domain, username),
        )
        return response

    def waiting(self) -> requests.Response:
        response = self._request(
            method="get",
            url=OnyxClient.ENDPOINTS["waiting"](self.config.domain),
        )
        return response

    def site_users(self) -> requests.Response:
        response = self._request(
            method="get",
            url=OnyxClient.ENDPOINTS["siteusers"](self.config.domain),
        )
        return response

    def all_users(self) -> requests.Response:
        response = self._request(
            method="get",
            url=OnyxClient.ENDPOINTS["allusers"](self.config.domain),
        )
        return response


def onyx_errors(method):
    """
    Decorator that coerces `requests` library errors into appropriate `OnyxError` subclasses.
    """
    if inspect.isgeneratorfunction(method):

        def wrapped_generator_method(self, *args, **kwargs):
            try:
                yield from method(self, *args, **kwargs)

            except HTTPError as e:
                if e.response is None:
                    # TODO: Seems this does not need handling?
                    raise e  #  type: ignore
                elif e.response.status_code < 500:
                    raise OnyxRequestError(
                        message=str(e),
                        response=e.response,
                    ) from e
                else:
                    raise OnyxServerError(
                        message=str(e),
                        response=e.response,
                    ) from e
            except RequestException as e:
                raise OnyxConnectionError(str(e)) from e

        return wrapped_generator_method
    else:

        def wrapped_method(self, *args, **kwargs):
            try:
                return method(self, *args, **kwargs)

            except HTTPError as e:
                if e.response is None:
                    # TODO: Seems this does not need handling?
                    raise e  #  type: ignore
                elif e.response.status_code < 500:
                    raise OnyxRequestError(
                        message=str(e),
                        response=e.response,
                    ) from e
                else:
                    raise OnyxServerError(
                        message=str(e),
                        response=e.response,
                    ) from e
            except RequestException as e:
                raise OnyxConnectionError(str(e)) from e

        return wrapped_method


class OnyxClient(OnyxClientBase):
    """
    Class for querying and manipulating data within Onyx.
    """

    def __init__(self, config: OnyxConfig) -> None:
        """
        Initialise a client.

        Args:
            config: `OnyxConfig` object that stores information for connecting and authenticating with Onyx.

        Examples:
            The recommended way to initialise a client (as a context manager):
            ```python
            import os
            from onyx import OnyxConfig, OnyxEnv, OnyxClient

            config = OnyxConfig(
                domain=os.environ[OnyxEnv.DOMAIN],
                token=os.environ[OnyxEnv.TOKEN],
            )

            with OnyxClient(config) as client:
                pass # Do something with the client here
            ```

            Alternatively, the client can be initialised as follows:
            ```python
            import os
            from onyx import OnyxConfig, OnyxEnv, OnyxClient

            config = OnyxConfig(
                domain=os.environ[OnyxEnv.DOMAIN],
                token=os.environ[OnyxEnv.TOKEN],
            )

            client = OnyxClient(config)
            # Do something with the client here
            ```

        Tips:
            - When making multiple requests, using the client as a context manager can improve performance.
            - This is due to the fact that the client will re-use the same session for all requests, rather than creating a new session for each request.
            - For more information, see: https://requests.readthedocs.io/en/master/user/advanced/#session-objects
        """
        super().__init__(config)

    @onyx_errors
    def projects(self) -> List[Dict[str, str]]:
        """
        View available projects.

        Returns:
            List of projects.

        Examples:
            ```python
            import os
            from onyx import OnyxConfig, OnyxEnv, OnyxClient

            config = OnyxConfig(
                domain=os.environ[OnyxEnv.DOMAIN],
                token=os.environ[OnyxEnv.TOKEN],
            )

            with OnyxClient(config) as client:
                projects = client.projects()
            ```
            ```python
            >>> projects
            [
                {
                    "project": "project_1",
                    "scope": "admin",
                    "actions": [
                        "get",
                        "list",
                        "filter",
                        "add",
                        "change",
                        "delete",
                    ],
                },
                {
                    "project": "project_2",
                    "scope": "analyst",
                    "actions": [
                        "get",
                        "list",
                        "filter",
                    ],
                },
            ]
            ```
        """

        response = super().projects()
        response.raise_for_status()
        return response.json()["data"]

    @onyx_errors
    def types(self) -> List[Dict[str, Any]]:
        """
        View available field types.

        Returns:
            List of field types.

        Examples:
            ```python
            import os
            from onyx import OnyxConfig, OnyxEnv, OnyxClient

            config = OnyxConfig(
                domain=os.environ[OnyxEnv.DOMAIN],
                token=os.environ[OnyxEnv.TOKEN],
            )

            with OnyxClient(config) as client:
                field_types = client.types()
            ```
            ```python
            >>> field_types
            [
                {
                    "type": "text",
                    "description": "A string of characters.",
                    "lookups": [
                        "exact",
                        "ne",
                        "in",
                        "notin",
                        "contains",
                        "startswith",
                        "endswith",
                        "iexact",
                        "icontains",
                        "istartswith",
                        "iendswith",
                        "length",
                        "length__in",
                        "length__range",
                        "isnull",
                    ],
                },
                {
                    "type": "choice",
                    "description": "A restricted set of options.",
                    "lookups": [
                        "exact",
                        "ne",
                        "in",
                        "notin",
                        "isnull",
                    ],
                },
            ]
            ```
        """

        response = super().types()
        response.raise_for_status()
        return response.json()["data"]

    @onyx_errors
    def lookups(self) -> List[Dict[str, Any]]:
        """
        View available lookups.

        Returns:
            List of lookups.

        Examples:
            ```python
            import os

            from onyx import OnyxConfig, OnyxEnv, OnyxClient

            config = OnyxConfig(
                domain=os.environ[OnyxEnv.DOMAIN],
                token=os.environ[OnyxEnv.TOKEN],
            )

            with OnyxClient(config) as client:
                lookups = client.lookups()
            ```
            ```python
            >>> lookups
            [
                {
                    "lookup": "exact",
                    "description": "The field's value must be equal to the query value.",
                    "types": [
                        "text",
                        "choice",
                        "integer",
                        "decimal",
                        "date",
                        "datetime",
                        "bool",
                    ],
                },
                {
                    "lookup": "ne",
                    "description": "The field's value must not be equal to the query value.",
                    "types": [
                        "text",
                        "choice",
                        "integer",
                        "decimal",
                        "date",
                        "datetime",
                        "bool",
                    ],
                },
            ]
            ```
        """

        response = super().lookups()
        response.raise_for_status()
        return response.json()["data"]

    @onyx_errors
    def fields(self, project: str) -> Dict[str, Any]:
        """
        View fields for a project.

        Args:
            project: Name of the project.

        Returns:
            Dict of fields.

        Examples:
            ```python
            import os
            from onyx import OnyxConfig, OnyxEnv, OnyxClient

            config = OnyxConfig(
                domain=os.environ[OnyxEnv.DOMAIN],
                token=os.environ[OnyxEnv.TOKEN],
            )

            with OnyxClient(config) as client:
                fields = client.fields("project")
            ```
            ```python
            >>> fields
            {
                "version": "0.1.0",
                "fields": {
                    "climb_id": {
                        "description": "Unique identifier for a project record in Onyx.",
                        "type": "text",
                        "required": True,
                        "actions": [
                            "get",
                            "list",
                            "filter",
                        ],
                        "restrictions": [
                            "Max length: 12",
                        ],
                    },
                    "is_published": {
                        "description": "Indicator for whether a project record has been published.",
                        "type": "bool",
                        "required": False,
                        "actions": [
                            "get",
                            "list",
                            "filter",
                            "add",
                            "change",
                        ],
                        "default": True,
                    },
                    "published_date": {
                        "description": "The date the project record was published in Onyx.",
                        "type": "date (YYYY-MM-DD)",
                        "required": False,
                        "actions": [
                            "get",
                            "list",
                            "filter",
                        ],
                    },
                    "country": {
                        "description": "Country of origin.",
                        "type": "choice",
                        "required": False,
                        "actions": [
                            "get",
                            "list",
                            "filter",
                            "add",
                            "change",
                        ],
                        "values": [
                            "ENG",
                            "WALES",
                            "SCOT",
                            "NI",
                        ],
                    },
                },
            }
            ```
        """

        response = super().fields(project)
        response.raise_for_status()
        return response.json()["data"]

    @onyx_errors
    def choices(self, project: str, field: str) -> Dict[str, Dict[str, Any]]:
        """
        View choices for a field.

        Args:
            project: Name of the project.
            field: Choice field on the project.

        Returns:
            Dictionary mapping choices to information about the choice.

        Examples:
            ```python
            import os
            from onyx import OnyxConfig, OnyxEnv, OnyxClient

            config = OnyxConfig(
                domain=os.environ[OnyxEnv.DOMAIN],
                token=os.environ[OnyxEnv.TOKEN],
            )

            with OnyxClient(config) as client:
                choices = client.choices("project", "country")
            ```
            ```python
            >>> choices
            {
                "ENG": {
                    "description": "England",
                    "is_active" : True,
                },
                "WALES": {
                    "description": "Wales",
                    "is_active" : True,
                },
                "SCOT": {
                    "description": "Scotland",
                    "is_active" : True,
                },
                "NI": {
                    "description": "Northern Ireland",
                    "is_active" : True,
                },
            }
            ```
        """

        response = super().choices(project, field)
        response.raise_for_status()
        return response.json()["data"]

    @onyx_errors
    def get(
        self,
        project: str,
        climb_id: Optional[str] = None,
        fields: Optional[Dict[str, Any]] = None,
        include: Union[List[str], str, None] = None,
        exclude: Union[List[str], str, None] = None,
    ) -> Dict[str, Any]:
        """
        Get a record from a project.

        Args:
            project: Name of the project.
            climb_id: Unique identifier for the record in the project.
            fields: Dictionary of field filters used to uniquely identify the record.
            include: Fields to include in the output.
            exclude: Fields to exclude from the output.

        Returns:
            Dict containing the record.

        Examples:
            Get a record by CLIMB ID:
            ```python
            import os
            from onyx import OnyxConfig, OnyxEnv, OnyxClient

            config = OnyxConfig(
                domain=os.environ[OnyxEnv.DOMAIN],
                token=os.environ[OnyxEnv.TOKEN],
            )

            with OnyxClient(config) as client:
                record = client.get("project", "C-1234567890")
            ```
            ```python
            >>> record
            {
                "climb_id": "C-1234567890",
                "published_date": "2023-01-01",
                "field1": "value1",
                "field2": "value2",
            }
            ```

            Get a record by fields that uniquely identify it:
            ```python
            import os
            from onyx import OnyxConfig, OnyxEnv, OnyxClient

            config = OnyxConfig(
                domain=os.environ[OnyxEnv.DOMAIN],
                token=os.environ[OnyxEnv.TOKEN],
            )

            with OnyxClient(config) as client:
                record = client.get(
                    "project",
                    fields={
                        "field1": "value1",
                        "field2": "value2",
                    },
                )
            ```
            ```python
            >>> record
            {
                "climb_id": "C-1234567890",
                "published_date": "2023-01-01",
                "field1": "value1",
                "field2": "value2",
            }
            ```

            The `include` and `exclude` arguments can be used to control the fields returned:
            ```python
            import os
            from onyx import OnyxConfig, OnyxEnv, OnyxClient

            config = OnyxConfig(
                domain=os.environ[OnyxEnv.DOMAIN],
                token=os.environ[OnyxEnv.TOKEN],
            )

            with OnyxClient(config) as client:
                record_v1 = client.get(
                    "project",
                    climb_id="C-1234567890",
                    include=["climb_id", "published_date"],
                )
                record_v2 = client.get(
                    "project",
                    climb_id="C-1234567890",
                    exclude=["field2"],
                )
            ```
            ```python
            >>> record_v1
            {
                "climb_id": "C-1234567890",
                "published_date": "2023-01-01",
            }
            >>> record_v2
            {
                "climb_id": "C-1234567890",
                "published_date": "2023-01-01",
                "field1": "value1",
            }
            ```

        Tips:
            - Including/excluding fields to reduce the size of the returned data can improve performance.
        """

        if climb_id and fields:
            raise OnyxClientError("Cannot provide both 'climb_id' and 'fields'.")

        if not (climb_id or fields):
            raise OnyxClientError("Must provide either 'climb_id' or 'fields'.")

        if climb_id:
            response = super().get(
                project,
                climb_id,
                include=include,
                exclude=exclude,
            )
            response.raise_for_status()
            return response.json()["data"]
        else:
            responses = super().filter(
                project,
                fields=fields,
                include=include,
                exclude=exclude,
            )
            response = next(responses, None)
            if response is None:
                raise OnyxClientError(
                    f"Expected one record to be returned but received no response."
                )

            response.raise_for_status()
            count = len(response.json()["data"])
            if count != 1:
                raise OnyxClientError(
                    f"Expected one record to be returned but received: {count}"
                )

            return response.json()["data"][0]

    @onyx_errors
    def filter(
        self,
        project: str,
        fields: Optional[Dict[str, Any]] = None,
        include: Union[List[str], str, None] = None,
        exclude: Union[List[str], str, None] = None,
        summarise: Union[List[str], str, None] = None,
        **kwargs: Any,
    ) -> Generator[Dict[str, Any], Any, None]:
        """
        Filter records from a project.

        Args:
            project: Name of the project.
            fields: Dictionary of field filters.
            include: Fields to include in the output.
            exclude: Fields to exclude from the output.
            summarise: For a given field (or group of fields), return the frequency of each unique value (or unique group of values).
            **kwargs: Additional keyword arguments are interpreted as field filters.

        Returns:
            Generator of records. If a summarise argument is provided, each record will be a dict containing values of the summary fields and a count for the frequency.

        Notes:
            - Field filters specify requirements that the returned data must satisfy. They can be provided as keyword arguments, or as a dictionary to the `fields` argument.
            - These filters can be a simple match on a value (e.g. `"published_date" : "2023-01-01"`), or they can use a 'lookup' for more complex matching conditions (e.g. `"published_date__iso_year" : "2023"`).
            - Multi-value lookups (e.g. `in`, `range`) can also be used. For keyword arguments, multiple values can be provided as a Python list. For the `fields` dictionary, multiple values must be provided as a comma-separated string (see examples below).

        Examples:
            Retrieve all records that match a set of field requirements:
            ```python
            import os
            from onyx import OnyxConfig, OnyxEnv, OnyxClient

            config = OnyxConfig(
                domain=os.environ[OnyxEnv.DOMAIN],
                token=os.environ[OnyxEnv.TOKEN],
            )

            # Field conditions can either be provided as keyword arguments:
            with OnyxClient(config) as client:
                records = list(
                    client.filter(
                        project="project",
                        field1="abcd",
                        published_date__range=["2023-01-01", "2023-01-02"],
                    )
                )

            # Or as a dictionary to the 'fields' argument:
            with OnyxClient(config) as client:
                records = list(
                    client.filter(
                        project="project",
                        fields={
                            "field1": "abcd",
                            "published_date__range" : "2023-01-01, 2023-01-02",
                        },
                    )
                )
            ```
            ```python
            >>> records
            [
                {
                    "climb_id": "C-1234567890",
                    "published_date": "2023-01-01",
                    "field1": "abcd",
                    "field2": 123,
                },
                {
                    "climb_id": "C-1234567891",
                    "published_date": "2023-01-02",
                    "field1": "abcd",
                    "field2": 456,
                },
            ]
            ```

            The `summarise` argument can be used to return the frequency of each unique value for a given field, or the frequency of each unique set of values for a group of fields:
            ```python
            import os
            from onyx import OnyxConfig, OnyxEnv, OnyxClient

            config = OnyxConfig(
                domain=os.environ[OnyxEnv.DOMAIN],
                token=os.environ[OnyxEnv.TOKEN],
            )

            with OnyxClient(config) as client:
                records_v1 = list(
                    client.filter(
                        project="project",
                        field1="abcd",
                        published_date__range=["2023-01-01", "2023-01-02"],
                        summarise="published_date",
                    )
                )

                records_v2 = list(
                    client.filter(
                        project="project",
                        field1="abcd",
                        published_date__range=["2023-01-01", "2023-01-02"],
                        summarise=["published_date", "field2"],
                    )
                )
            ```
            ```python
            >>> records_v1
            [
                {
                    "published_date": "2023-01-01",
                    "count": 1,
                },
                {
                    "published_date": "2023-01-02",
                    "count": 1,
                },
            ]
            >>> records_v2
            [
                {
                    "published_date": "2023-01-01",
                    "field2": 123,
                    "count": 1,
                },
                {
                    "published_date": "2023-01-02",
                    "field2": 456,
                    "count": 1,
                },
            ]
            ```

        """

        responses = super().filter(
            project,
            fields=fields,
            include=include,
            exclude=exclude,
            summarise=summarise,
            **kwargs,
        )
        for response in responses:
            response.raise_for_status()
            for result in response.json()["data"]:
                yield result

    @onyx_errors
    def query(
        self,
        project: str,
        query: Optional[OnyxField] = None,
        include: Union[List[str], str, None] = None,
        exclude: Union[List[str], str, None] = None,
        summarise: Union[List[str], str, None] = None,
    ) -> Generator[Dict[str, Any], Any, None]:
        """
        Query records from a project.

        This method supports more complex filtering than the `OnyxClient.filter` method.
        Here, filters can be combined using Python's bitwise operators, representing AND, OR, XOR and NOT operations.

        Args:
            project: Name of the project.
            query: `OnyxField` object representing the query being made.
            include: Fields to include in the output.
            exclude: Fields to exclude from the output.
            summarise: For a given field (or group of fields), return the frequency of each unique value (or unique group of values).

        Returns:
            Generator of records. If a summarise argument is provided, each record will be a dict containing values of the summary fields and a count for the frequency.

        Notes:
            - The `query` argument must be an instance of `OnyxField`.
            - `OnyxField` instances can be combined into complex expressions using Python's bitwise operators: `&` (AND), `|` (OR), `^` (XOR), and `~` (NOT).
            - Multi-value lookups (e.g. `in`, `range`) support passing a Python list (see example below).

        Examples:
            Retrieve all records that match the query provided by an `OnyxField` object:
            ```python
            import os
            from onyx import OnyxConfig, OnyxEnv, OnyxClient, OnyxField

            config = OnyxConfig(
                domain=os.environ[OnyxEnv.DOMAIN],
                token=os.environ[OnyxEnv.TOKEN],
            )

            with OnyxClient(config) as client:
                records = list(
                    client.query(
                        project="project",
                        query=(
                            OnyxField(field1="abcd")
                            & OnyxField(published_date__range=["2023-01-01", "2023-01-02"])
                        ),
                    )
                )
            ```
            ```python
            >>> records
            [
                {
                    "climb_id": "C-1234567890",
                    "published_date": "2023-01-01",
                    "field1": "abcd",
                    "field2": 123,
                },
                {
                    "climb_id": "C-1234567891",
                    "published_date": "2023-01-02",
                    "field1": "abcd",
                    "field2": 456,
                },
            ]
            ```
        """

        responses = super().query(
            project,
            query=query,
            include=include,
            exclude=exclude,
            summarise=summarise,
        )
        for response in responses:
            response.raise_for_status()
            for result in response.json()["data"]:
                yield result

    @classmethod
    @onyx_errors
    def to_csv(
        cls,
        csv_file: TextIO,
        data: Union[List[Dict[str, Any]], Generator[Dict[str, Any], Any, None]],
        delimiter: Optional[str] = None,
    ):
        """
        Write a set of records to a CSV file.

        Args:
            csv_file: File object for the CSV file being written to.
            data: The data being written to the CSV file. Must be either a list / generator of dict records.
            delimiter: CSV delimiter. If not provided, defaults to `","` for CSVs. Set this to `"\\t"` to work with TSV files.

        Examples:
            ```python
            import os
            from onyx import OnyxConfig, OnyxEnv, OnyxClient

            config = OnyxConfig(
                domain=os.environ[OnyxEnv.DOMAIN],
                token=os.environ[OnyxEnv.TOKEN],
            )

            with OnyxClient(config) as client, open("/path/to/file.csv") as csv_file:
                client.to_csv(
                    csv_file=csv_file,
                    data=client.filter(
                        "project",
                        fields={
                            "field1": "value1",
                            "field2": "value2",
                        },
                )
            ```
        """
        super().to_csv(
            csv_file=csv_file,
            data=data,
            delimiter=delimiter,
        )

    @onyx_errors
    def history(
        self,
        project: str,
        climb_id: str,
    ) -> Dict[str, Any]:
        """
        View the history of a record in a project.

        Args:
            project: Name of the project.
            climb_id: Unique identifier for the record in the project.

        Returns:
            Dict containing the history of the record.

        Examples:
            ```python
            import os
            from onyx import OnyxConfig, OnyxEnv, OnyxClient

            config = OnyxConfig(
                domain=os.environ[OnyxEnv.DOMAIN],
                token=os.environ[OnyxEnv.TOKEN],
            )

            with OnyxClient(config) as client:
                history = client.history("project", "C-1234567890")
            ```
            ```python
            >>> history
            {
                "climb_id": "C-1234567890",
                "history": [
                    {
                        "username": "user",
                        "timestamp": "2023-01-01T00:00:00Z",
                        "action": "add",
                    },
                    {
                        "username": "user",
                        "timestamp": "2023-01-02T00:00:00Z",
                        "action": "change",
                        "changes": [
                            {
                                "field": "field_1",
                                "type": "text",
                                "from": "value1",
                                "to": "value2",
                            },
                            {
                                "field": "field_2",
                                "type": "integer",
                                "from": 3,
                                "to": 4,
                            },
                            {
                                "field": "nested_field",
                                "type": "relation",
                                "action": "add",
                                "count" : 3,
                            },
                            {
                                "field": "nested_field",
                                "type": "relation",
                                "action": "change",
                                "count" : 10,
                            },
                        ],
                    },
                ],
            }
            ```
        """
        response = super().history(project, climb_id)
        response.raise_for_status()
        return response.json()["data"]

    @onyx_errors
    def identify(
        self,
        project: str,
        field: str,
        value: str,
        site: Optional[str] = None,
    ) -> Dict[str, str]:
        """
        Get the anonymised identifier for a value on a field.

        Args:
            project: Name of the project.
            field: Field on the project.
            value: Value to identify.
            site: Site to identify the value on. If not provided, defaults to the user's site.

        Returns:
            Dict containing the project, site, field, value and anonymised identifier.

        Examples:
            ```python
            import os
            from onyx import OnyxConfig, OnyxEnv, OnyxClient

            config = OnyxConfig(
                domain=os.environ[OnyxEnv.DOMAIN],
                token=os.environ[OnyxEnv.TOKEN],
            )

            with OnyxClient(config) as client:
                identification = client.identify("project", "sample_id", "hidden-value")
            ```
            ```python
            >>> identification
            {
                "project": "project",
                "site": "site",
                "field": "sample_id",
                "value": "hidden-value",
                "identifier": "S-1234567890",
            }
            ```
        """

        response = super().identify(project, field, value, site=site)
        response.raise_for_status()
        return response.json()["data"]

    @onyx_errors
    def create(
        self,
        project: str,
        fields: Dict[str, Any],
        test: bool = False,
    ) -> Dict[str, Any]:
        """
        Create a record in a project.

        Args:
            project: Name of the project.
            fields: Object representing the record to be created.
            test: If `True`, runs the command as a test. Default: `False`

        Returns:
            Dict containing the CLIMB ID of the created record.

        Examples:
            ```python
            import os
            from onyx import OnyxConfig, OnyxEnv, OnyxClient

            config = OnyxConfig(
                domain=os.environ[OnyxEnv.DOMAIN],
                token=os.environ[OnyxEnv.TOKEN],
            )

            with OnyxClient(config) as client:
                result = client.create(
                    "project",
                    fields={
                        "field1": "value1",
                        "field2": "value2",
                    },
                )
            ```
            ```python
            >>> result
            {"climb_id": "C-1234567890"}
            ```
        """

        response = super().create(project, fields, test=test)
        response.raise_for_status()
        return response.json()["data"]

    @onyx_errors
    def update(
        self,
        project: str,
        climb_id: str,
        fields: Optional[Dict[str, Any]] = None,
        test: bool = False,
    ) -> Dict[str, Any]:
        """
        Update a record in a project.

        Args:
            project: Name of the project.
            climb_id: Unique identifier for the record in the project.
            fields: Object representing the record to be updated.
            test: If `True`, runs the command as a test. Default: `False`

        Returns:
            Dict containing the CLIMB ID of the updated record.

        Examples:
            ```python
            import os
            from onyx import OnyxConfig, OnyxEnv, OnyxClient

            config = OnyxConfig(
                domain=os.environ[OnyxEnv.DOMAIN],
                token=os.environ[OnyxEnv.TOKEN],
            )

            with OnyxClient(config) as client:
                result = client.update(
                    project="project",
                    climb_id="C-1234567890",
                    fields={
                        "field1": "value1",
                        "field2": "value2",
                    },
                )
            ```
            ```python
            >>> result
            {"climb_id": "C-1234567890"}
            ```
        """

        response = super().update(project, climb_id, fields=fields, test=test)
        response.raise_for_status()
        return response.json()["data"]

    @onyx_errors
    def delete(
        self,
        project: str,
        climb_id: str,
    ) -> Dict[str, Any]:
        """
        Delete a record in a project.

        Args:
            project: Name of the project.
            climb_id: Unique identifier for the record in the project.

        Returns:
            Dict containing the CLIMB ID of the deleted record.

        Examples:
            ```python
            import os
            from onyx import OnyxConfig, OnyxEnv, OnyxClient

            config = OnyxConfig(
                domain=os.environ[OnyxEnv.DOMAIN],
                token=os.environ[OnyxEnv.TOKEN],
            )

            with OnyxClient(config) as client:
                result = client.delete(
                    project="project",
                    climb_id="C-1234567890",
                )
            ```
            ```python
            >>> result
            {"climb_id": "C-1234567890"}
            ```
        """

        response = super().delete(project, climb_id)
        response.raise_for_status()
        return response.json()["data"]

    @onyx_errors
    def csv_create(
        self,
        project: str,
        csv_file: TextIO,
        fields: Optional[Dict[str, Any]] = None,
        delimiter: Optional[str] = None,
        multiline: bool = False,
        test: bool = False,
    ) -> Union[Dict[str, Any], List[Dict[str, Any]]]:
        """
        Use a CSV file to create record(s) in a project.

        Args:
            project: Name of the project.
            csv_file: File object for the CSV file being used for record upload.
            fields: Additional fields provided for each record being uploaded. Takes precedence over fields in the CSV.
            delimiter: CSV delimiter. If not provided, defaults to `","` for CSVs. Set this to `"\\t"` to work with TSV files.
            multiline: If `True`, allows processing of CSV files with more than one record. Default: `False`
            test: If `True`, runs the command as a test. Default: `False`

        Returns:
            Dict containing the CLIMB ID of the created record. If `multiline = True`, returns a list of dicts containing the CLIMB ID of each created record.

        Examples:
            Create a single record:
            ```python
            import os
            from onyx import OnyxConfig, OnyxEnv, OnyxClient

            config = OnyxConfig(
                domain=os.environ[OnyxEnv.DOMAIN],
                token=os.environ[OnyxEnv.TOKEN],
            )

            with OnyxClient(config) as client, open("/path/to/file.csv") as csv_file:
                result = client.csv_create(
                    project="project",
                    csv_file=csv_file,
                )
            ```
            ```python
            >>> result
            {"climb_id": "C-1234567890"}
            ```

            Create multiple records:
            ```python
            import os
            from onyx import OnyxConfig, OnyxEnv, OnyxClient

            config = OnyxConfig(
                domain=os.environ[OnyxEnv.DOMAIN],
                token=os.environ[OnyxEnv.TOKEN],
            )

            with OnyxClient(config) as client, open("/path/to/file.csv") as csv_file:
                results = client.csv_create(
                    project="project",
                    csv_file=csv_file,
                    multiline=True,
                )
            ```
            ```python
            >>> results
            [
                {"climb_id": "C-1234567890"},
                {"climb_id": "C-1234567891"},
                {"climb_id": "C-1234567892"},
            ]
            ```
        """

        responses = super().csv_create(
            project,
            csv_file=csv_file,
            fields=fields,
            delimiter=delimiter,
            multiline=multiline,
            test=test,
        )
        return self._csv_handle_multiline(responses, multiline)

    @onyx_errors
    def csv_update(
        self,
        project: str,
        csv_file: TextIO,
        fields: Optional[Dict[str, Any]] = None,
        delimiter: Optional[str] = None,
        multiline: bool = False,
        test: bool = False,
    ) -> Union[Dict[str, Any], List[Dict[str, Any]]]:
        """
        Use a CSV file to update record(s) in a project.

        Args:
            project: Name of the project.
            csv_file: File object for the CSV file being used for record upload.
            fields: Additional fields provided for each record being uploaded. Takes precedence over fields in the CSV.
            delimiter: CSV delimiter. If not provided, defaults to `","` for CSVs. Set this to `"\\t"` to work with TSV files.
            multiline: If `True`, allows processing of CSV files with more than one record. Default: `False`
            test: If `True`, runs the command as a test. Default: `False`

        Returns:
            Dict containing the CLIMB ID of the updated record. If `multiline = True`, returns a list of dicts containing the CLIMB ID of each updated record.

        Examples:
            Update a single record:
            ```python
            import os
            from onyx import OnyxConfig, OnyxEnv, OnyxClient

            config = OnyxConfig(
                domain=os.environ[OnyxEnv.DOMAIN],
                token=os.environ[OnyxEnv.TOKEN],
            )

            with OnyxClient(config) as client, open("/path/to/file.csv") as csv_file:
                result = client.csv_update(
                    project="project",
                    csv_file=csv_file,
                )
            ```
            ```python
            >>> result
            {"climb_id": "C-1234567890"}
            ```

            Update multiple records:
            ```python
            import os
            from onyx import OnyxConfig, OnyxEnv, OnyxClient

            config = OnyxConfig(
                domain=os.environ[OnyxEnv.DOMAIN],
                token=os.environ[OnyxEnv.TOKEN],
            )

            with OnyxClient(config) as client, open("/path/to/file.csv") as csv_file:
                results = client.csv_update(
                    project="project",
                    csv_file=csv_file,
                    multiline=True,
                )
            ```
            ```python
            >>> results
            [
                {"climb_id": "C-1234567890"},
                {"climb_id": "C-1234567891"},
                {"climb_id": "C-1234567892"},
            ]
            ```
        """

        responses = super().csv_update(
            project,
            csv_file=csv_file,
            fields=fields,
            delimiter=delimiter,
            multiline=multiline,
            test=test,
        )
        return self._csv_handle_multiline(responses, multiline)

    @onyx_errors
    def csv_delete(
        self,
        project: str,
        csv_file: TextIO,
        delimiter: Optional[str] = None,
        multiline: bool = False,
    ) -> Union[Dict[str, Any], List[Dict[str, Any]]]:
        """
        Use a CSV file to delete record(s) in a project.

        Args:
            project: Name of the project.
            csv_file: File object for the CSV file being used for record upload.
            delimiter: CSV delimiter. If not provided, defaults to `","` for CSVs. Set this to `"\\t"` to work with TSV files.
            multiline: If `True`, allows processing of CSV files with more than one record. Default: `False`

        Returns:
            Dict containing the CLIMB ID of the deleted record. If `multiline = True`, returns a list of dicts containing the CLIMB ID of each deleted record.

        Examples:
            Delete a single record:
            ```python
            import os
            from onyx import OnyxConfig, OnyxEnv, OnyxClient

            config = OnyxConfig(
                domain=os.environ[OnyxEnv.DOMAIN],
                token=os.environ[OnyxEnv.TOKEN],
            )

            with OnyxClient(config) as client, open("/path/to/file.csv") as csv_file:
                result = client.csv_delete(
                    project="project",
                    csv_file=csv_file,
                )
            ```
            ```python
            >>> result
            {"climb_id": "C-1234567890"}
            ```

            Delete multiple records:
            ```python
            import os
            from onyx import OnyxConfig, OnyxEnv, OnyxClient

            config = OnyxConfig(
                domain=os.environ[OnyxEnv.DOMAIN],
                token=os.environ[OnyxEnv.TOKEN],
            )

            with OnyxClient(config) as client, open("/path/to/file.csv") as csv_file:
                results = client.csv_delete(
                    project="project",
                    csv_file=csv_file,
                    multiline=True,
                )
            ```
            ```python
            >>> results
            [
                {"climb_id": "C-1234567890"},
                {"climb_id": "C-1234567891"},
                {"climb_id": "C-1234567892"},
            ]
            ```
        """

        responses = super().csv_delete(
            project,
            csv_file=csv_file,
            delimiter=delimiter,
            multiline=multiline,
        )
        return self._csv_handle_multiline(responses, multiline)

    @classmethod
    @onyx_errors
    def register(
        cls,
        domain: str,
        first_name: str,
        last_name: str,
        email: str,
        site: str,
        password: str,
    ) -> Dict[str, Any]:
        """
        Create a new user.

        Args:
            domain: Name of the domain.
            first_name: First name of the user.
            last_name: Last name of the user.
            email: Email address of the user.
            site: Name of the site.
            password: Password for the user.

        Returns:
            Dict containing the user's information.

        Examples:
            ```python
            import os
            from onyx import OnyxClient, OnyxEnv

            registration = OnyxClient.register(
                domain=os.environ[OnyxEnv.DOMAIN],
                first_name="Bill",
                last_name="Will",
                email="bill@email.com",
                site="site",
                password="pass123",
            )
            ```
            ```python
            >>> registration
            {
                "username": "onyx-willb",
                "site": "site",
                "email": "bill@email.com",
                "first_name": "Bill",
                "last_name": "Will",
            }
            ```
        """
        response = super().register(
            domain,
            first_name,
            last_name,
            email,
            site,
            password,
        )
        response.raise_for_status()
        return response.json()["data"]

    @onyx_errors
    def login(self) -> Dict[str, Any]:
        """
        Log in the user.

        Returns:
            Dict containing the user's authentication token and it's expiry.

        Examples:
            ```python
            import os
            from onyx import OnyxConfig, OnyxEnv, OnyxClient

            config = OnyxConfig(
                domain=os.environ[OnyxEnv.DOMAIN],
                username=os.environ[OnyxEnv.USERNAME],
                password=os.environ[OnyxEnv.PASSWORD],
            )

            with OnyxClient(config) as client:
                token = client.login()
            ```
            ```python
            >>> token
            {
                "expiry": "2024-01-01T00:00:00.000000Z",
                "token": "abc123",
            }
            ```
        """

        response = super().login()
        response.raise_for_status()
        return response.json()["data"]

    @onyx_errors
    def logout(self) -> None:
        """
        Log out the user.

        Examples:
            ```python
            import os
            from onyx import OnyxConfig, OnyxEnv, OnyxClient

            config = OnyxConfig(
                domain=os.environ[OnyxEnv.DOMAIN],
                token=os.environ[OnyxEnv.TOKEN],
            )

            with OnyxClient(config) as client:
                client.logout()
            ```
        """

        response = super().logout()
        response.raise_for_status()

    @onyx_errors
    def logoutall(self) -> None:
        """
        Log out the user in all clients.

        Examples:
            ```python
            import os
            from onyx import OnyxConfig, OnyxEnv, OnyxClient

            config = OnyxConfig(
                domain=os.environ[OnyxEnv.DOMAIN],
                token=os.environ[OnyxEnv.TOKEN],
            )

            with OnyxClient(config) as client:
                client.logoutall()
            ```
        """

        response = super().logoutall()
        response.raise_for_status()

    @onyx_errors
    def profile(self) -> Dict[str, str]:
        """
        View the user's information.

        Returns:
            Dict containing the user's information.

        Examples:
            ```python
            import os
            from onyx import OnyxConfig, OnyxEnv, OnyxClient

            config = OnyxConfig(
                domain=os.environ[OnyxEnv.DOMAIN],
                token=os.environ[OnyxEnv.TOKEN],
            )

            with OnyxClient(config) as client:
                profile = client.profile()
            ```
            ```python
            >>> profile
            {
                "username": "user",
                "site": "site",
                "email": "user@email.com",
            }
            ```
        """

        response = super().profile()
        response.raise_for_status()
        return response.json()["data"]

    @onyx_errors
    def activity(self) -> List[Dict[str, Any]]:
        """
        View the user's latest activity.

        Returns:
            List of the user's latest activity.

        Examples:
            ```python
            import os
            from onyx import OnyxConfig, OnyxEnv, OnyxClient

            config = OnyxConfig(
                domain=os.environ[OnyxEnv.DOMAIN],
                token=os.environ[OnyxEnv.TOKEN],
            )

            with OnyxClient(config) as client:
                activity = client.activity()
            ```
            ```python
            >>> activity
            [
                {
                    "date": "2023-01-01T00:00:00.000000Z",
                    "address": "127.0.0.1",
                    "endpoint": "/projects/project/",
                    "method": "POST",
                    "status": 400,
                    "exec_time": 29,
                    "error_messages" : "b'{\"status\":\"fail\",\"code\":400,\"messages\":{\"site\":[\"Select a valid choice.\"]}}'",
                },
                {
                    "timestamp": "2023-01-02T00:00:00.000000Z",
                    "address": "127.0.0.1",
                    "endpoint": "/accounts/activity/",
                    "method": "GET",
                    "status": 200,
                    "exec_time": 22,
                    "error_messages": "",
                },
            ]
            ```
        """

        response = super().activity()
        response.raise_for_status()
        return response.json()["data"]

    @onyx_errors
    def approve(self, username: str) -> Dict[str, Any]:
        """
        Approve another user.

        Args:
            username: Username of the user to be approved.

        Returns:
            Dict confirming user approval success.

        Examples:
            ```python
            import os
            from onyx import OnyxConfig, OnyxEnv, OnyxClient

            config = OnyxConfig(
                domain=os.environ[OnyxEnv.DOMAIN],
                token=os.environ[OnyxEnv.TOKEN],
            )

            with OnyxClient(config) as client:
                approval = client.approve("waiting_user")
            ```
            ```python
            >>> approval
            {
                "username": "waiting_user",
                "is_approved": True,
            }
            ```
        """

        response = super().approve(username)
        response.raise_for_status()
        return response.json()["data"]

    @onyx_errors
    def waiting(self) -> List[Dict[str, Any]]:
        """
        Get users waiting for approval.

        Returns:
            List of users waiting for approval.

        Examples:
            ```python
            import os
            from onyx import OnyxConfig, OnyxEnv, OnyxClient

            config = OnyxConfig(
                domain=os.environ[OnyxEnv.DOMAIN],
                token=os.environ[OnyxEnv.TOKEN],
            )

            with OnyxClient(config) as client:
                users = client.waiting()
            ```
            ```python
            >>> users
            [
                {
                    "username": "waiting_user",
                    "site": "site",
                    "email": "waiting_user@email.com",
                    "date_joined": "2023-01-01T00:00:00.000000Z",
                }
            ]
            ```
        """

        response = super().waiting()
        response.raise_for_status()
        return response.json()["data"]

    @onyx_errors
    def site_users(self) -> List[Dict[str, Any]]:
        """
        Get users within the site of the requesting user.

        Returns:
            List of users within the site of the requesting user.

        Examples:
            ```python
            import os
            from onyx import OnyxConfig, OnyxEnv, OnyxClient

            config = OnyxConfig(
                domain=os.environ[OnyxEnv.DOMAIN],
                token=os.environ[OnyxEnv.TOKEN],
            )

            with OnyxClient(config):
                users = client.site_users()
            ```
            ```python
            >>> users
            [
                {
                    "username": "user",
                    "site": "site",
                    "email": "user@email.com",
                }
            ]
            ```
        """

        response = super().site_users()
        response.raise_for_status()
        return response.json()["data"]

    @onyx_errors
    def all_users(self) -> List[Dict[str, Any]]:
        """
        Get all users.

        Returns:
            List of all users.

        Examples:
            ```python
            import os
            from onyx import OnyxConfig, OnyxEnv, OnyxClient

            config = OnyxConfig(
                domain=os.environ[OnyxEnv.DOMAIN],
                token=os.environ[OnyxEnv.TOKEN],
            )

            with OnyxClient(config) as client:
                users = client.all_users()
            ```
            ```python
            >>> users
            [
                {
                    "username": "user",
                    "site": "site",
                    "email": "user@email.com",
                },
                {
                    "username": "another_user",
                    "site": "another_site",
                    "email": "another_user@email.com",
                },
            ]
            ```
        """

        response = super().all_users()
        response.raise_for_status()
        return response.json()["data"]



from __future__ import annotations
from typing import Any
from .exceptions import OnyxFieldError


class OnyxOperator:
    AND = "&"
    OR = "|"
    XOR = "^"
    NOT = "~"


class OnyxField:
    """
    Class that represents a single field-value pair for use in Onyx queries.
    """

    __slots__ = "query"

    def __init__(self, **kwargs: Any) -> None:
        """
        Initialise a field.

        Args:
            **kwargs: Keyword arguments containing a single key-value pair.

        Notes:
            - Takes a single key-value argument as input.
            - The key corresponds to a field (and optional lookup) to use for filtering.
            - The value corresponds to the field value(s) that are being matched against.
            - `OnyxField` instances can be combined into complex expressions using Python's bitwise operators: `&` (AND), `|` (OR), `^` (XOR), and `~` (NOT).
            - Multi-value lookups (e.g. `in`, `range`) support passing a Python list as the value. These are coerced into comma-separated strings internally.

        Examples:
            Create `OnyxField` objects and combine them using Python bitwise operators:
            ```python
            from onyx import OnyxField

            field1 = OnyxField(field1="value1")
            field2 = OnyxField(field2__contains="value2")

            expression = (field1 | field2) & OnyxField(
                published_date__range=["2023-01-01", "2023-01-02"]
            )
            ```
            ```python
            >>> field1
            <onyx.field.OnyxField object at 0x1028eb850>
            >>> field2
            <onyx.field.OnyxField object at 0x1028eb850>
            >>> expression
            <onyx.field.OnyxField object at 0x103b6fc40>
            >>> field1.query
            {"field1": "value1"}
            >>> field2.query
            {"field2__contains": "value2"}
            >>> expression.query
            {
                "&": [
                    {"|": [{"field1": "value1"}, {"field2__contains": "value2"}]},
                    {"published_date__range": "2023-01-01,2023-01-02"},
                ]
            }
            ```
        """

        if len(kwargs) != 1:
            raise OnyxFieldError(
                f"Expected exactly one field-value pair as a keyword argument. Received: {len(kwargs)}"
            )

        # Get the field-value pair from the kwargs
        field, value = next(iter(kwargs.items()))

        # If the field is not an operation and it is a multi-value lookup
        # Join the values into a comma-separated string
        if field not in {
            OnyxOperator.AND,
            OnyxOperator.OR,
            OnyxOperator.XOR,
            OnyxOperator.NOT,
        }:
            if type(value) in {list, tuple, set}:
                value = ",".join(map(lambda x: str(x) if x is not None else "", value))

        self.query = {field: value}

    def _validate_field(self, field: OnyxField) -> None:
        """
        Ensure an instance with the correct type has been provided.
        """
        if not isinstance(field, OnyxField):
            raise OnyxFieldError(
                f"Expected another instance of the same type. Received: {type(field)}"
            )

    def _combine_on_associative(self, field: OnyxField, operation: str) -> OnyxField:
        """
        Combine two pre-existing queries on an ASSOCIATIVE operation (`AND`, `OR`, `XOR`), and use this to reduce nested JSON.
        E.g. take the following query:

        `((X AND Y) AND Z) OR (W AND (X AND (Y OR (Z OR X))))`

        We can use associativity of `AND` and `OR` to reduce how deeply nested the JSON request body is.
        By associativity of `AND` and `OR`, the following is unambiguous and logically equivalent to the previous query:

        `(X AND Y AND Z) OR (W AND X AND (Y OR Z OR X))`

        With the former corresponding to more deeply-nested JSON than the latter.
        This is useful! There is normally a limit on how deeply nested a JSON request body can be.
        So by preventing unnecessary nesting, users can programatically construct and execute a broader range of queries.
        """

        # For each field, if the topmost key is equal to the current operation, we pull up the values.
        # Otherwise, they stay self-contained within their existing operation.
        # For example, if operation the operation is '&' and we have self = {"&" : [{...}]} and field = {"|" : [{...}]}
        # Then this function would take [{...}] from self, and create [{"|" : [{...}]}] from field
        # And then return {"&" : [{...}] + [{"|" : [{...}]}]} or {"&" : [{...}, {"|" : [{...}]}]}
        self_key, self_value = next(iter(self.query.items()))
        if self_key == operation:
            if not isinstance(self_value, list):
                raise OnyxFieldError(
                    f"Expected associative operation key '{self_key}' to point to a list. Received: {type(self_value)}"
                )
            self_query = self_value
        else:
            self_query = [self.query]

        field_key, field_value = next(iter(field.query.items()))
        if field_key == operation:
            if not isinstance(field_value, list):
                raise OnyxFieldError(
                    f"Expected associative operation key '{field_key}' to point to a list. Received: {type(self_value)}"
                )
            field_query = field_value
        else:
            field_query = [field.query]

        return OnyxField(**{operation: self_query + field_query})

    def __eq__(self, field: OnyxField) -> bool:
        self._validate_field(field)
        return self.query == field.query

    def __and__(self, field: OnyxField) -> OnyxField:
        self._validate_field(field)
        return self._combine_on_associative(field, OnyxOperator.AND)

    def __or__(self, field: OnyxField) -> OnyxField:
        self._validate_field(field)
        return self._combine_on_associative(field, OnyxOperator.OR)

    def __xor__(self, field: OnyxField) -> OnyxField:
        self._validate_field(field)
        return self._combine_on_associative(field, OnyxOperator.XOR)

    def __invert__(self) -> OnyxField:
        # Here we account for double negatives to also reduce nesting
        # Not really needed as people are (unlikely) to be putting multiple negations one after the other
        # But hey you never know

        # Get the top-most key of the current query
        # If its also a NOT, we pull out the value and initialise that as the query
        self_key, self_value = next(iter(self.query.items()))
        if self_key == OnyxOperator.NOT:
            if not isinstance(self_value, dict):
                raise OnyxFieldError(
                    f"Expected operation key '{self_key}' to point to a dict. Received: {type(self_value)}"
                )

            return OnyxField(**self_value)
        else:
            return OnyxField(**{OnyxOperator.NOT: self.query})
